import sqlalchemy as sa
import sqlalchemy.orm as so
from typing import Optional

from ...db import Base

class Vocabulary(Base): 
    __tablename__ = 'vocabulary'
    vocabulary_id : so.Mapped[str] = so.mapped_column(sa.String(20), primary_key=True)
    vocabulary_name: so.Mapped[str] = so.mapped_column(sa.String(255))
    vocabulary_reference: so.Mapped[Optional[str]] = so.mapped_column(sa.String(255), nullable=True)
    vocabulary_version: so.Mapped[Optional[str]] = so.mapped_column(sa.String(255), nullable=True)
    vocabulary_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='v_fk_1'))
    
    #concept: so.Mapped['Concept'] = so.relationship(foreign_keys=[vocabulary_concept_id], back_populates='concept_vocabulary')
    vocabulary_concept: so.Mapped['Concept'] = so.relationship('Concept', primaryjoin='Vocabulary.vocabulary_concept_id==Concept.concept_id')




import sqlalchemy as sa
import sqlalchemy.orm as so

from omop_alchemy.cdm.base import (
    Base,
    CDMTableBase,
    cdm_table, 
    required_concept_fk,
)

@cdm_table
class Fact_Relationship(CDMTableBase, Base):
    __tablename__ = "fact_relationship"

    domain_concept_id_1: so.Mapped[int] = required_concept_fk()
    fact_id_1: so.Mapped[int] = so.mapped_column(sa.Integer,nullable=False)
    domain_concept_id_2: so.Mapped[int] = required_concept_fk()
    fact_id_2: so.Mapped[int] = so.mapped_column(sa.Integer,nullable=False)
    relationship_concept_id: so.Mapped[int] = required_concept_fk()

    __table_args__ = (
        sa.PrimaryKeyConstraint(
            "domain_concept_id_1",
            "fact_id_1",
            "domain_concept_id_2",
            "fact_id_2",
            "relationship_concept_id",
        ),
    )

    def __repr__(self) -> str:
        return (
            "<FactRelationship "
            f"{self.domain_concept_id_1}:{self.fact_id_1} "
            f"-[{self.relationship_concept_id}]-> "
            f"{self.domain_concept_id_2}:{self.fact_id_2}>"
        )


from datetime import date, datetime, time
from typing import Optional, List
from decimal import Decimal
import sqlalchemy as sa
import sqlalchemy.orm as so
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy

from ...db import Base
from ...conventions.concept_enumerators import ModifierFields

class Modifiable_Table(Base):
    __tablename__ = 'modifiable_table'
    modifier_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True)
    modifier_of_field_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='mt_fk_1'))

    modifiers: so.Mapped[List['Measurement']] = so.relationship(
        backref="modifying_object", lazy="selectin", viewonly=True
    )

    related_obs: so.Mapped[List['Observation']] = so.relationship(
        backref="observing_object", lazy="selectin", viewonly=True
    )

    related_events: so.Mapped[List['Episode_Event']] = so.relationship(
        backref="event_object", lazy="selectin", viewonly=True
    )

    modifier_concepts: AssociationProxy[List[int]] = association_proxy("modifiers", "measurement_concept_id")
    modifier_value_concepts: AssociationProxy[List[int]] = association_proxy("modifiers", "value_as_concept_id")

    @property
    def primary_event(self):
        return [event.primary_ep for event in self.related_events]
    
    @property
    def event_date(self):
        return None


    __mapper_args__ = {
            "polymorphic_on":sa.case(
                (modifier_of_field_concept_id == ModifierFields.condition_occurrence_id.value, "condition"),
                (modifier_of_field_concept_id == ModifierFields.procedure_occurrence_id.value, "procedure"),
                (modifier_of_field_concept_id == ModifierFields.drug_exposure_id.value, "drug_exposure"),
                else_="episode"),
            "polymorphic_identity":"measurement"
        }
    
    
    @property
    def polymorphic_label(self):
        return self.__mapper_args__['polymorphic_identity']


from datetime import datetime, date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Visit_Occurrence(Base):
    __tablename__ = 'visit_occurrence'
    # identifier
    visit_occurrence_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True) 
    # temporal
    visit_start_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    visit_start_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    visit_end_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    visit_end_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    # strings
    visit_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    admitted_from_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    discharge_to_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    # numeric
    admitted_from_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer, nullable=True)
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='vo_fk_1'))
    provider_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('provider.provider_id', name='vo_fk_2'), nullable=True)
    care_site_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('care_site.care_site_id', name='vo_fk_3'), nullable=True)
    preceding_visit_occurrence_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_occurrence.visit_occurrence_id', name='vo_fk_4'), nullable=True)
    # concept fks
    visit_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='vo_fk_5'))
    visit_source_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='vo_fk_6'), nullable=True)
    discharge_to_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='vo_fk_7'), nullable=True)
    visit_type_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='vo_fk_8'))
    # relationships
    person: so.Mapped[Optional['Person']] = so.relationship(foreign_keys=[person_id])
    provider: so.Mapped[Optional['Provider']] = so.relationship(foreign_keys=[provider_id])
    care_site: so.Mapped[Optional['Care_Site']] = so.relationship(foreign_keys=[care_site_id])
    visit_occurrence: so.Mapped[Optional['Visit_Occurrence']] = so.relationship(foreign_keys=[preceding_visit_occurrence_id])
    # concept_relationships
    visit_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[visit_concept_id])
    visit_source: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[visit_source_concept_id])
    discharge_to: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[discharge_to_concept_id])
    visit_type: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[visit_type_concept_id])


import sqlalchemy as sa
import sqlalchemy.orm as so
from ...db import Base


class Relationship(Base): 
    __tablename__ = 'relationship'
    relationship_id: so.Mapped[str] = so.mapped_column(sa.String(20), primary_key=True)
    relationship_name: so.Mapped[str] = so.mapped_column(sa.String(255))
    is_hierarchical: so.Mapped[str]  = so.mapped_column(sa.String(1))
    defines_ancestry: so.Mapped[str]  = so.mapped_column(sa.String(1))
    reverse_relationship_id: so.Mapped[str] = so.mapped_column(sa.ForeignKey('relationship.relationship_id', name='r_fk_2'))
    relationship_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='r_fk_1'))

    concept: so.Mapped['Concept'] = so.relationship('Concept', primaryjoin='Relationship.relationship_concept_id==Concept.concept_id')
    reverse: so.Mapped['Relationship'] = so.relationship(foreign_keys=[reverse_relationship_id])

    def __repr__(self):
        return f'<Relationship {self.relationship_id} - {self.relationship_name}>'
    




# from .condition_occurrence import Condition_Occurrence
# from .device_exposure import Device_Exposure
# from .drug_exposure import Drug_Exposure
# from .fact_relationship import Fact_Relationship
# from .measurement import Measurement
# from .note_nlp import Note_NLP
# from .note import Note
# from .observation_period import Observation_Period
# from .observation import Observation
# from .person import Person
# from .procedure_occurrence import Procedure_Occurrence
# from .specimen import Specimen
# from .survey_conduct import Survey_Conduct
# from .visit_occurrence import Visit_Occurrence
# from .visit_detail import Visit_Detail
# from .modifiable_table import Modifiable_Table
# from .death import Death

# for concept_linked_table in [Observation, Measurement]:
#     concept_linked_table.add_concepts()

# __all__ = [Condition_Occurrence, Device_Exposure, Drug_Exposure, Fact_Relationship, Measurement,
#            Note_NLP, Note, Observation_Period, Observation, Person, Procedure_Occurrence, Specimen, 
#            Survey_Conduct, Visit_Occurrence, Visit_Detail, Modifiable_Table, Death]


from datetime import datetime, date
from typing import Optional, List
from sqlalchemy.ext.hybrid import hybrid_property
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Death(Base):
    __tablename__ = 'death'
    
    # we need a primary key here, so we enforce a single death record per person on a given day although this is not
    # per CDM definition - to follow up - may want to add death key transparently instead?
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='dth_fk_1'), primary_key=True)

    death_date: so.Mapped[date] = so.mapped_column(sa.Date, primary_key=True)
    death_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime, nullable=True)
    
    cause_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    
    death_type_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='dth_fk_2'), nullable=True)
    cause_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='dth_fk_3'), nullable=True)
    cause_source_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='dth_fk_4'), nullable=True)
    
    death_type: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[death_type_concept_id])
    cause: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[cause_concept_id])
    cause_source: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[cause_source_concept_id])

    person: so.Mapped['Person'] = so.relationship(foreign_keys=[person_id])



from datetime import datetime, date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so
from sqlalchemy.ext.hybrid import hybrid_property
from .modifiable_table import Modifiable_Table
from ..vocabulary.vocabulary import Concept
from ...db import Base
from ...conventions.concept_enumerators import ModifierFields

class Drug_Exposure(Modifiable_Table):
    __tablename__ = 'drug_exposure'    
    # identifier
    drug_exposure_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('modifiable_table.modifier_id'), primary_key=True, autoincrement=True)

    # temporal
    drug_exposure_start_date: so.Mapped[date] = so.mapped_column(sa.Date)
    drug_exposure_start_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    drug_exposure_end_date: so.Mapped[date] = so.mapped_column(sa.Date)
    drug_exposure_end_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)    
    verbatim_end_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date, nullable=True)
    # strings
    stop_reason: so.Mapped[Optional[str]] = so.mapped_column(sa.String(20), nullable=True)
    lot_number: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    drug_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    route_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    dose_unit_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    sig: so.Mapped[Optional[str]] = so.mapped_column(sa.Text, nullable=True)
    # numeric
    refills: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer, nullable=True)
    days_supply: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer, nullable=True)
    quantity: so.Mapped[Optional[int]] = so.mapped_column(sa.Numeric, nullable=True)
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='dr_fk_1'))
    provider_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('provider.provider_id', name='dr_fk_2'), nullable=True)
    visit_occurrence_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_occurrence.visit_occurrence_id', name='dr_fk_3'), nullable=True)
    visit_detail_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_detail.visit_detail_id', name='dr_fk_4'), nullable=True)
    # concept fks
    drug_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='dr_fk_5'))
    drug_type_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='dr_fk_6'))
    route_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='dr_fk_7'), nullable=True)
    drug_source_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='dr_fk_8'), nullable=True)
    # relationships
    person: so.Mapped['Person'] = so.relationship(foreign_keys=[person_id])
    provider: so.Mapped[Optional['Provider']] = so.relationship(foreign_keys=[provider_id])
    visit_occurrence: so.Mapped[Optional['Visit_Occurrence']] = so.relationship(foreign_keys=[visit_occurrence_id])
    visit_detail: so.Mapped[Optional['Visit_Detail']] = so.relationship(foreign_keys=[visit_detail_id])
    # concept_relationships
    drug_concept: so.Mapped['Concept'] = so.relationship(foreign_keys=[drug_concept_id])
    drug_type_concept: so.Mapped['Concept'] = so.relationship(foreign_keys=[drug_type_concept_id])
    route_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[route_concept_id])
    drug_source_concept: so.Mapped['Concept'] = so.relationship(foreign_keys=[drug_source_concept_id])


    __mapper_args__ = {
        "polymorphic_identity": "drug_exposure",
        'inherit_condition': (drug_exposure_id == Modifiable_Table.modifier_id)
    }

    def __init__(self, *args, **kwargs):
        super().__init__(modifier_of_field_concept_id = ModifierFields.drug_exposure_id.value, *args, **kwargs)

    @property
    def event_date(self):
        return  self.drug_exposure_start_datetime.date() if self.drug_exposure_start_datetime is not None else self.drug_exposure_start_date

    @property
    def drug_label(self):
        if self.drug_concept:
            return self.drug_concept.concept_name
    
    @property
    def route_label(self):
        if self.route_concept:
            return self.route_concept.concept_name
    


# from .episode import Episode
# from .episode_event import Episode_Event

# for concept_linked_table in [Episode]:
#     concept_linked_table.add_concepts()
# # from .joined_mappers import Condition_Episode, Systemic_Therapy_Episode, Historical_Surgical_Procedure, Dated_Surgical_Procedure,\
# #                             Radiation_Therapy_Episode, Person_Episodes, Dx_Treat_Start, Dx_RT_Start, Dx_SACT_Start


# __all__ = [Episode, Episode_Event]


# # , Condition_Episode, Systemic_Therapy_Episode, Radiation_Therapy_Episode, Historical_Surgical_Procedure, 
# #            Dated_Surgical_Procedure, Person_Episodes, Dx_Treat_Start, Dx_RT_Start, Dx_SACT_Start]


from datetime import datetime
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Fact_Relationship(Base):
    __tablename__ = 'fact_relationship'
    # identifier
    fact_relationship_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True) 
    # temporal
    # strings
    # numeric
    fact_id_1: so.Mapped[int] = so.mapped_column(sa.Integer)
    fact_id_2: so.Mapped[int] = so.mapped_column(sa.Integer)
    # fks
    # concept fks
    domain_concept_id_1: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='fr_fk_1'))
    domain_concept_id_2: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='fr_fk_2'))
    relationship_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='fr_fk_3'))
    # relationships
    # concept_relationships
    relationship: so.Mapped['Concept'] = so.relationship(foreign_keys=[relationship_concept_id])
    domain_1: so.Mapped['Concept'] = so.relationship(foreign_keys=[domain_concept_id_1])
    domain_2: so.Mapped['Concept'] = so.relationship(foreign_keys=[domain_concept_id_2])



from datetime import datetime
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Cohort_Definition(Base):
    __tablename__ = 'cohort_definition'
    cohort_definition_id: so.Mapped[int] = so.mapped_column(index=True, unique=True, primary_key=True)



from datetime import date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so
from ...db import Base

class Concept_Relationship(Base): 
    __tablename__ = 'concept_relationship'

    concept_id_1: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='cr_fk_1'), primary_key=True)
    concept_id_2: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='cr_fk_2'), primary_key=True)
    relationship_id: so.Mapped[str] = so.mapped_column(sa.String(20), sa.ForeignKey('relationship.relationship_id', name='cr_fk_3'), primary_key=True)

    valid_start_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    valid_end_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    invalid_reason: so.Mapped[Optional[str]]  = so.mapped_column(sa.String(1), nullable=True)

    concept_1: so.Mapped['Concept'] = so.relationship('Concept', primaryjoin='Concept_Relationship.concept_id_1==Concept.concept_id', back_populates='concept_relationships')
    concept_2: so.Mapped['Concept'] = so.relationship('Concept', primaryjoin='Concept_Relationship.concept_id_2==Concept.concept_id')
    relationship: so.Mapped['Relationship'] = so.relationship(foreign_keys=[relationship_id])

    def __repr__(self):
        return f'<Relationship {self.concept_1.concept_name} ({self.relationship.relationship_id}) {self.concept_2.concept_name}>'


from datetime import date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Observation_Period(Base):
    __tablename__ = 'observation_period'
    # identifier
    observation_period_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True) 
    # temporal
    observation_period_start_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    observation_period_end_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    # strings
    # numeric
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='op_fk_1'))
    # concept fks
    period_type_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='op_fk_2'))
    # relationships
    person: so.Mapped[Optional['Person']] = so.relationship(foreign_keys=[person_id])
    # concept_relationships
    period_type: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[period_type_concept_id])





from typing import Any
from sqlalchemy import Integer, Float, Boolean, Date, DateTime, Enum, String, Text
from typing import Any
from datetime import date, datetime 
import math


_ATHENA_DATE_FORMATS = (
    "%d-%b-%Y",        # 24-AUG-2017 (Athena standard)
    "%Y-%m-%d",        # ISO (sometimes appears)
    "%d/%m/%Y",        # defensive
)

def _parse_date(value: str) -> date | None:
    for fmt in _ATHENA_DATE_FORMATS:
        try:
            return datetime.strptime(value, fmt).date()
        except ValueError:
            continue
    return None


def _parse_datetime(value: str) -> datetime | None:
    # Try datetime first
    try:
        return datetime.fromisoformat(value)
    except ValueError:
        pass

    # Fallback to date-only formats + midnight
    d = _parse_date(value)
    if d:
        return datetime.combine(d, datetime.min.time())

    return None

def _to_bool(value: Any) -> bool | None:
    if value is None or (isinstance(value, float) and math.isnan(value)):
        return None
    if isinstance(value, bool):
        return value
    s = str(value).strip().lower()
    if s in {"true", "t", "yes", "y", "1"}:
        return True
    if s in {"false", "f", "no", "n", "0"}:
        return False
    return None

def perform_cast(value: Any, col_type: Any) -> Any:
    if value is None:
        return None
    if isinstance(value, float) and math.isnan(value):
        return None
    if isinstance(value, str) and value.strip() == "":
        return None
    # Integer
    if type(col_type) == Integer:
        try:
            return int(value)
        except (ValueError, TypeError):
            return None

    # Float
    if type(col_type) == Float:
        try:
            return float(value)
        except (ValueError, TypeError):
            return None
    # Boolean
    if type(col_type) == Boolean:
        return _to_bool(value)

    # Date, DateTime
    if type(col_type) == Date:
        if isinstance(value, date) and not isinstance(value, datetime):
            return value
        if isinstance(value, datetime):
            return value.date()
        if isinstance(value, str):
            return _parse_date(value)
        return None
    
    if type(col_type) == DateTime:
        if isinstance(value, datetime):
            return value
        if isinstance(value, date):
            return datetime.combine(value, datetime.min.time())
        if isinstance(value, str):
            return _parse_datetime(value)
        return None

    # String / Text
    if type(col_type) == String or type(col_type) == Text:
        # Canonicalise numeric-looking identifiers
        if isinstance(value, float) and value.is_integer():
            return str(int(value))

        if isinstance(value, str):
            v = value.strip()
            if v == "":
                return None
            if v.isdigit():
                return str(int(v))
            try:
                f = float(v)
                if f.is_integer():
                    return str(int(f))
            except ValueError:
                pass
            return v

        return str(value)
    # Fallback: leave as is
    return value



from .declarative import Base, create_db, bootstrap, get_table_by_name
from .typing import HasTableName, ORMTable, DomainSemanticTable
from .cdm_constants import ModifierFieldConcepts
from .decorators import cdm_table
from .mixins import (
    PersonScoped,
    ConceptTyped,
    DatedEvent,
    ValueMixin,
    HealthSystemContext,
    FactTable,
    ReferenceTable,
    SourceAttribution,
    UnitConcept,
    DomainValidationMixin,
    ExpectedDomain,
)
from .modifiable import ModifierTargetMixin, ModifierTargetRegistry
from .reference_context import ReferenceContextMixin
from .column_helpers import required_concept_fk, optional_concept_fk, optional_fk, required_int, optional_int
from .cdm_table_base import CDMTableBase

__all__ = [
    "Base", 
    "HasTableName",
    "ModifierFieldConcepts",
    "cdm_table",
    "PersonScoped",
    "ConceptTyped",
    "DatedEvent",
    "ValueMixin",
    "HealthSystemContext",
    "FactTable",
    "ReferenceTable",
    "SourceAttribution",
    "UnitConcept",
    "ReferenceContextMixin",
    "required_concept_fk",
    "optional_concept_fk",
    "optional_fk",
    "required_int",
    "optional_int",
    "CDMTableBase",
    "DomainValidationMixin",
    "ExpectedDomain",
    "ORMTable",
    "DomainSemanticTable",
    "create_db",
    "ModifierTargetMixin",
    "ModifierTargetRegistry",
    "bootstrap",
    "get_table_by_name"
]


import sqlalchemy as sa
import sqlalchemy.orm as so
from sqlalchemy.ext.declarative import declared_attr
from typing import Optional, TYPE_CHECKING, List
from datetime import date
from sqlalchemy.ext.hybrid import hybrid_method
from sqlalchemy.orm.exc import DetachedInstanceError
from functools import cached_property
from omop_alchemy.cdm.base import (
    Base, 
    cdm_table,
    CDMTableBase, 
    required_concept_fk,
    optional_concept_fk,
    PersonScoped,
    optional_int,
    ReferenceContextMixin,
    DomainValidationMixin,
    ExpectedDomain,
)

from ...base.declarative import get_table_by_name

if TYPE_CHECKING:
    from ..vocabulary import Concept
    from .episode import Episode

@cdm_table
class Episode_Event(CDMTableBase, Base):
    __tablename__ = "episode_event"

    episode_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey("episode.episode_id"),nullable=False,primary_key=True)
    event_id: so.Mapped[int] = so.mapped_column(nullable=False,primary_key=True)
    episode_event_field_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey("concept.concept_id"),nullable=False,primary_key=True)

    def __repr__(self) -> str:
        return f"<EpisodeEvent ep={self.episode_id} event={self.event_id}>"
    
class Episode_EventContext(ReferenceContextMixin):
    episode: so.Mapped["Episode"] = ReferenceContextMixin._reference_relationship(target="Episode",local_fk="episode_id",remote_pk="episode_id",)  # type: ignore[assignment]
    event_field: so.Mapped["Concept"] = ReferenceContextMixin._reference_relationship(target="Concept",local_fk="episode_event_field_concept_id",remote_pk="concept_id",)  # type: ignore[assignment]

class Episode_EventView(Episode_Event, Episode_EventContext, DomainValidationMixin):
    """
    Episode ↔ Event linkage view.

    Identifies which CDM table the EVENT_ID comes from via
    episode_event_field_concept_id.
    """

    __tablename__ = "episode_event"
    __mapper_args__ = {"concrete": False}

    __expected_domains__ = {
        "episode_event_field_concept_id": ExpectedDomain("Metadata"),
    }


    @property
    def event_table(self) -> str | None:
        if self.event_field and "." in self.event_field.concept_name:
            return self.event_field.concept_name.split(".", 1)[0]
        return None

    @cached_property
    def resolved_event(self):
        """
        Resolve EVENT_ID to concrete OMOP row.
        Cached per-instance.
        """
        table_name = self.event_table
        session = so.object_session(self)
        if session is None or table_name is None:
            return None

        cls = get_table_by_name(table_name)
        if cls is not None:
            return session.get(cls, self.event_id) # type: ignore
        return None

    def __repr__(self):
        target = self.resolved_event
        if target is not None:
            return (
                f"<EpisodeEvent ep={self.episode_id} "
                f"{target.__class__.__name__}#{self.event_id}>"
            )
        return f"<EpisodeEvent ep={self.episode_id} event={self.event_id}>"
    
    @property
    def episode_start_datetime(self):
        return (
            self.episode.episode_start_datetime
            if self.episode else None
        )
    


import pandas as pd
import json
from datetime import date, datetime
from typing import Type
import sqlalchemy as sa
import sqlalchemy.orm as so
from .typing import HasTableName
from ..utils import CDMValidationError, perform_cast, get_logger

logger = get_logger(__name__)


def _json_default(obj):
    if isinstance(obj, (date, datetime)):
        return obj.isoformat()
    raise TypeError(f"Not JSON serializable: {type(obj)}")


def iter_chunks(df: pd.DataFrame, size: int):
    for start in range(0, len(df), size):
        yield df.iloc[start : start + size]

def load_by_chunk(
    cls: Type[HasTableName],
    session: so.Session,
    dataframe: pd.DataFrame,
    chunk_size: int = 10_000) -> int:

    total = 0

    for chunk in iter_chunks(dataframe, chunk_size):
        records = chunk.to_dict(orient="records")
        objs = [cls(**r) for r in records]  # type: ignore - assume we are only getting to this point with valid data
        session.add_all(objs)
        session.flush()
        total += len(objs)

    return total

def normalise_csv_to_model(model_columns: dict[str, sa.Column], df: pd.DataFrame, cls: Type[HasTableName]) -> pd.DataFrame:
    if df.empty:
        return df
    
    for col, sa_col in model_columns.items():
        logger.debug(f'Coercing data from csv against model for {cls.__tablename__}.{col}')
        if col not in df.columns:
            continue
        df[col] = df[col].map(lambda v: perform_cast(v, sa_col.type))

    required_cols = [name for name, col in model_columns.items() if not col.nullable]
    null_mask = df[required_cols].isna().any(axis=1)
    null_count = int(null_mask.sum())
    if null_count > 0:
        logger.warning("Found %d rows with NULL primary keys in %s", null_count, cls.__tablename__)
        df = df.loc[~null_mask]
    return df
                
def dedupe_csv(df: pd.DataFrame, pk_names: list[str], cls: Type[HasTableName], session: so.Session, max_bind_vars=1_000) -> pd.DataFrame:
    before = len(df)
    df = df.drop_duplicates(subset=pk_names, keep="first")
    dropped = before - len(df)
    logger.info(f'Dropping {dropped} duplicates from csv file for table {cls.__tablename__}')
    pk_tuples = list(df[pk_names].itertuples(index=False, name=None))
    pk_cols = [getattr(cls, c) for c in pk_names]

    vars_per_row = len(pk_cols)
    chunk_size = max_bind_vars // vars_per_row
    existing_rows: list[tuple] = []
    for i in range(0, len(pk_tuples), chunk_size):
        chunk = pk_tuples[i : i + chunk_size]

        rows = (
            session.query(*pk_cols)
            .filter(sa.tuple_(*pk_cols).in_(chunk))
            .all()
        )

        existing_rows.extend(rows)


    if not existing_rows:
        return df

    existing = pd.DataFrame(existing_rows, columns=pk_names)


    # existing = pd.DataFrame(
    #     session.query(*pk_cols)
    #     .filter(sa.tuple_(*pk_cols).in_(pk_tuples))
    # )
    if len(existing) > 0:
        logger.warning(f'{len(existing)} duplicate records in csv file {cls.__tablename__} will be dropped')
        df = df.merge(
            existing,
            on=pk_names,
            how="left",
            indicator=True,
        )
        df = df[df["_merge"] == "left_only"].drop(columns="_merge")
    return df


import sqlalchemy.orm as so
import sqlalchemy as sa
import pandas as pd
import json
import hashlib
from pathlib import Path
from typing import Type, Any, Tuple, cast

from ..registry import ValidationIssue
from ..specification import TableSpec, FieldSpec
from ..utils import CDMValidationError, get_logger
from .file_helpers import load_by_chunk, normalise_csv_to_model, dedupe_csv, _json_default
from .typing import HasTableName

logger = get_logger(__name__)

class IdAllocator:
    """
    This class is used instead of database sequences, due to requirement for supporting
    SQLite which does not have sequence support.

    ID allocation assumes a single writer per table.
    Concurrent writers will fail fast with PK violations.
    """
    def __init__(self, start: int):
        self._next = start + 1

    def next(self) -> int:
        val = self._next
        self._next += 1
        return val

    def reserve(self, n: int) -> range:
        start = self._next
        self._next += n
        return range(start, start + n)


class CDMTableBase:
    """
    Base class for CDM tables that support CSV loading and validation.
    """
    __abstract__ = True

    """
    Adds structural validation of ORM models against
    the official OMOP CDM CSV specifications.
    """
    __cdm_extra_checks__: list[str] = []

    @classmethod
    def mapper_for(cls: Type) -> so.Mapper:
        return cast(so.Mapper, sa.inspect(cls))

    @classmethod
    def pk_columns(cls) -> list[sa.Column]:
        pks = list(cls.mapper_for().primary_key)
        if not pks:
            raise ValueError(f"{cls.__name__} has no primary key")
        return pks # type: ignore
    
    @classmethod
    def pk_values(cls, obj: Any) -> dict[str, Any]:
        """
        Return primary key values as a dict:
        {"person_id": 123, "visit_occurrence_id": 456}
        """
        return {
            col.key: getattr(obj, col.key)
            for col in cls.pk_columns()
        }

    @classmethod
    def pk_tuple(cls, obj: Any) -> Tuple[Any, ...]:
        """
        Return PK values as a positional tuple in PK order.
        Suitable for:
        - set membership
        - dedupe
        - tuple_(*cols).in_(...)
        """
        return tuple(
            getattr(obj, col.key)
            for col in cls.pk_columns()
        )

    @classmethod
    def pk_names(cls) -> list[str]:
        return [c.key for c in cls.pk_columns()]

    @classmethod
    def allocator(cls, session) -> IdAllocator:
        start = cls.max_id(session)
        return IdAllocator(start)

    @classmethod
    def max_id(cls, session) -> int:
        pks = cls.pk_columns()
        if len(pks) != 1:
            raise ValueError(f"{cls.__name__} has composite PK; max_id() not supported")
        pk = pks[0]
        return session.query(sa.func.max(pk)).scalar() or 0

    @classmethod
    def next_id(cls, session) -> int:
        return cls.max_id(session) + 1

    @classmethod
    def load_csv(
        cls: Type[HasTableName],
        session: so.Session,
        path: Path,
        *,
        strict: bool = True,
        delimiter: str = "\t",
        dedupe: bool = False,
        normalise: bool = True,
        chunk_size: int = 10_000,
    ) -> int:
        
        logger.debug(f'Loading csv file for {cls.__tablename__}')
        if path.stem.lower() != cls.__tablename__:
            raise ValueError(f"CSV filename '{path.name}' does not match table '{cls.__tablename__}'")

        mapper = cls.mapper_for()
        model_columns: dict[str, sa.Column] = {c.key: c for c in mapper.columns}
        pk_names = cls.pk_names()
        df = pd.read_csv(path, delimiter=delimiter, dtype=str)

        if strict:
            logger.debug(f'Checking csv file against model for {cls.__tablename__}')
            unknown = set(df.columns) - set(model_columns)
            if unknown:
                raise ValueError(f"Unknown columns in file: {cls.__tablename__}: {unknown}")
            missing = set(model_columns) - set(df.columns)
            if missing:
                raise ValueError(f"Missing columns in {cls.__tablename__}: {missing}")

        df = df[[c for c in df.columns if c in model_columns]]

        if normalise:
            df = normalise_csv_to_model(model_columns, df, cls)
        if dedupe:
            df = dedupe_csv(df, pk_names, cls, session)

        total = load_by_chunk(
            cls=cls,
            session=session,
            dataframe=df,
            chunk_size=chunk_size,
        )
        logger.info("%s: inserted %d rows from CSV",cls.__tablename__,total)
        return total

    @classmethod
    def extra_validate(cls) -> list["ValidationIssue"]:
        return []

    @classmethod
    def validate_against_spec(
        cls: Type[HasTableName],
        *,
        table_spec: TableSpec,
        field_specs: dict[str, FieldSpec],
    ) -> None:
        if cls.__tablename__ != table_spec.table_name:
            raise CDMValidationError(
                f"{cls.__name__}: table name mismatch "
                f"({cls.__tablename__} != {table_spec.table_name})"
            )

        mapper = sa.inspect(cls)

        if not mapper:
            raise CDMValidationError(f"{cls.__name__} is not a mapped ORM class")
    
        orm_columns = {c.key.lower() for c in mapper.columns}

        # Required fields must exist
        for fname, spec in field_specs.items():
            if spec.is_required and fname not in orm_columns:
                raise CDMValidationError(
                    f"{cls.__tablename__}: missing required column '{fname}'"
                )
            
    
    def to_dict(
        self,
        *,
        include_nulls: bool = False,
        only: set[str] | None = None,
        exclude: set[str] | None = None,
    ) -> dict[str, Any]:
        mapper = self.mapper_for()

        data = {}
        for col in mapper.columns:
            key = col.key
            if only and key not in only:
                continue
            if exclude and key in exclude:
                continue

            value = getattr(self, key)
            if value is None and not include_nulls:
                continue

            data[key] = value

        return data
    

    def to_json(self, **kwargs) -> str:
        return json.dumps(
            self.to_dict(**kwargs),
            default=_json_default,
            sort_keys=True,
        )
    
    def row_fingerprint(self) -> str:
        payload = self.to_json(include_nulls=True)
        return hashlib.sha256(payload.encode("utf-8")).hexdigest()


import sqlalchemy as sa
import sqlalchemy.orm as so
from typing import Optional
from datetime import date

from omop_alchemy.cdm.base import (
    Base,
    cdm_table,
    CDMTableBase,
    required_concept_fk,
)

@cdm_table
class CDM_Source(CDMTableBase, Base):
    __tablename__ = "cdm_source"

    cdm_source_name: so.Mapped[str] = so.mapped_column(sa.String(255), nullable=False, primary_key=True)
    cdm_source_abbreviation: so.Mapped[str] = so.mapped_column(sa.String(25), nullable=False)
    cdm_holder: so.Mapped[str] = so.mapped_column(sa.String(255), nullable=False)

    source_description: so.Mapped[Optional[str]] = so.mapped_column(sa.Text, nullable=True)
    source_documentation_reference: so.Mapped[Optional[str]] = so.mapped_column(sa.String(255), nullable=True)
    cdm_etl_reference: so.Mapped[Optional[str]] = so.mapped_column(sa.String(255), nullable=True)

    source_release_date: so.Mapped[date] = so.mapped_column(sa.Date, nullable=False)
    cdm_release_date: so.Mapped[date] = so.mapped_column(sa.Date, nullable=False)

    cdm_version: so.Mapped[Optional[str]] = so.mapped_column(sa.String(10), nullable=True)
    cdm_version_concept_id: so.Mapped[int] = required_concept_fk()

    vocabulary_version: so.Mapped[str] = so.mapped_column(sa.String(20), nullable=False)

    def __repr__(self) -> str:
        return f"<CDM_Source {self.cdm_source_abbreviation}>"



from __future__ import annotations
from importlib.resources import files
from importlib.resources.abc import Traversable

CDM_VERSION: str = "5.4"

_SPEC_ROOT: Traversable = files("omop_alchemy.cdm.specification")

TABLE_LEVEL_CSV: Traversable = _SPEC_ROOT / "OMOP_CDMv5.4_Table_Level.csv"
FIELD_LEVEL_CSV: Traversable = _SPEC_ROOT / "OMOP_CDMv5.4_Field_Level.csv"


import sqlalchemy as sa
import sqlalchemy.orm as so
from ...db import Base

class Domain(Base): 
    __tablename__ = 'domain'
    domain_id: so.Mapped[str] = so.mapped_column(sa.String(20), primary_key=True)
    domain_name: so.Mapped[str] = so.mapped_column(sa.String(255))
    domain_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='d_fk_1'))
    domain_concept: so.Mapped['Concept'] = so.relationship('Concept', primaryjoin='Domain.domain_concept_id==Concept.concept_id')
#    concept: so.Mapped['Concept'] = so.relationship(foreign_keys=[domain_concept_id])

    def __repr__(self):
        return f'<Domain {self.domain_id} - {self.domain_name}>'


import sqlalchemy.orm as so
import sqlalchemy as sa
from typing import TYPE_CHECKING, Type, Optional

from omop_alchemy.cdm.base.declarative import Base
from omop_alchemy.cdm.base.typing import HasTableName

if TYPE_CHECKING:
    from .registry import ValidationIssue

CDM_TO_SA = {
    "integer": {"integer", "biginteger"},
    "float": {"float", "numeric", "decimal", "real"},
    "date": {"date"},
    "datetime": {"datetime"},
    "varchar": {"string", "varchar", "unicode", "unicodetext", "text"}
}

def normalize_omop_type(raw: str) -> str:
    raw = raw.lower()
    if raw.startswith("varchar"):
        return "varchar"
    if raw in {"integer", "int"}:
        return "integer"
    return raw

def _sa_type_name(coltype: sa.types.TypeEngine) -> str:
    return coltype.__class__.__name__.lower()

def is_type_compatible(cdm_datatype: str, sa_col: sa.Column) -> bool:
    dt = cdm_datatype.strip().lower()
    sa_name = _sa_type_name(sa_col.type)

    # Normalize varchar(N) to generic varchar
    if dt.startswith("varchar(") and dt not in CDM_TO_SA:
        dt = "varchar"

    allowed = CDM_TO_SA.get(dt)
    if not allowed:
        # unknown in map: treat as warning elsewhere
        return True
    return sa_name in allowed

def varchar_length(sa_col: sa.Column) -> Optional[int]:
    if hasattr(sa_col.type, "length"):
        return sa_col.type.length # type: ignore
    return None

def validate_primary_key(table: sa.Table, spec_pk_fields: set[str]) -> tuple[bool, set[str], set[str]]:
    actual_pk = {c.name for c in table.primary_key.columns}
    missing = spec_pk_fields - actual_pk
    extra = actual_pk - spec_pk_fields
    ok = not missing and not extra
    return ok, missing, extra

def fk_targets(sa_col: sa.Column) -> set[tuple[str, str]]:
    targets: set[tuple[str, str]] = set()
    for fk in sa_col.foreign_keys:
        # fk.target_fullname like "concept.concept_id"
        t = fk.target_fullname.split(".")
        if len(t) == 2:
            targets.add((t[0], t[1]))
    return targets



from datetime import datetime, date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Note_NLP(Base):
    __tablename__ = 'note_nlp'
    # identifier
    note_nlp_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True) 
    # temporal
    nlp_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    nlp_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    # strings
    snippet: so.Mapped[Optional[str]] = so.mapped_column(sa.String(250), nullable=True)
    offset: so.Mapped[Optional[str]] = so.mapped_column(sa.String(250), nullable=True)
    lexical_variant: so.Mapped[Optional[str]] = so.mapped_column(sa.String(250), nullable=True)
    nlp_system: so.Mapped[Optional[str]] = so.mapped_column(sa.String(250), nullable=True)
    term_temporal: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    term_modifiers: so.Mapped[Optional[str]] = so.mapped_column(sa.String(2000), nullable=True)
    term_exists: so.Mapped[Optional[str]] = so.mapped_column(sa.String(1), nullable=True)
    # numeric
    # fks    
    note_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('note.note_id'))
    # concept fks
    section_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='nn_fk_1'), nullable=True)
    note_nlp_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='nn_fk_2'), nullable=True)
    note_nlp_source_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='nn_fk_3'), nullable=True)
    # relationships
    note: so.Mapped[Optional['Note']] = so.relationship(foreign_keys=[note_id])
    # concept_relationships
    section: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[section_concept_id])
    note_nlp: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[note_nlp_concept_id])
    note_nlp_source: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[note_nlp_source_concept_id])







from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Concept_Ancestor(Base): 
    __tablename__ = 'concept_ancestor'

    ancestor_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='ca_fk_1'), primary_key=True)
    descendant_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='ca_fk_2'), primary_key=True)

    ancestor: so.Mapped['Concept'] = so.relationship('Concept', primaryjoin='Concept_Ancestor.ancestor_concept_id==Concept.concept_id')
    descendant: so.Mapped['Concept'] = so.relationship('Concept', primaryjoin='Concept_Ancestor.descendant_concept_id==Concept.concept_id')
    min_levels_of_separation: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer)
    max_levels_of_separation: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer)

    def __repr__(self):
        return f'<Ancestry {self.ancestor.concept_name} ({self.min_levels_of_separation} - {self.max_levels_of_separation}) {self.descendant.concept_name}>'




from datetime import datetime, date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Note(Base):
    __tablename__ = 'note'
    # identifier
    note_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True) 
    # temporal
    note_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    note_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    # strings
    note_title: so.Mapped[Optional[str]] = so.mapped_column(sa.String(250), nullable=True)
    note_text: so.Mapped[Optional[str]] = so.mapped_column(sa.Text, nullable=True)
    note_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    # numeric
    note_event_id: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer, nullable=True)
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='n_fk_1'))
    provider_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('provider.provider_id', name='n_fk_2'), nullable=True)
    visit_occurrence_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_occurrence.visit_occurrence_id', name='n_fk_3'), nullable=True)
    visit_detail_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_detail.visit_detail_id', name='n_fk_4'), nullable=True)
    # concept fks
    note_type_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='n_fk_5'))
    note_class_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='n_fk_6'))
    encoding_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='n_fk_7'))
    language_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='n_fk_8'))
    note_event_field_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='n_fk_9'), nullable=True)
    # relationships
    person: so.Mapped[Optional['Person']] = so.relationship(foreign_keys=[person_id])
    provider: so.Mapped[Optional['Provider']] = so.relationship(foreign_keys=[provider_id])
    visit_occurrence: so.Mapped[Optional['Visit_Occurrence']] = so.relationship(foreign_keys=[visit_occurrence_id])
    visit_detail: so.Mapped[Optional['Visit_Detail']] = so.relationship(foreign_keys=[visit_detail_id])
    # concept_relationships
    note_type: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[note_type_concept_id])
    note_class: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[note_class_concept_id])
    encoding: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[encoding_concept_id])
    language: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[language_concept_id])
    note_event_field: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[note_event_field_concept_id])


from typing import Type, TypeVar
from .cdm_table_base import CDMTableBase

T = TypeVar("T", bound=type)

def cdm_table(cls: T) -> T:
    """
    Mark a SQLAlchemy declarative class as a concrete OMOP CDM table.

    - Forces __abstract__ = False
    - Ensures __tablename__ is defined
    - Inherits from CDMTableBase
    - Used to clearly distinguish real CDM tables from mixins
    """

    # Safety: require an explicit tablename
    if not hasattr(cls, "__tablename__"):
        raise TypeError(
            f"@cdm_table applied to {cls.__name__} "
            "but __tablename__ is not defined"
        )

    if not issubclass(cls, CDMTableBase):
        raise TypeError(
            f"{cls.__name__} must inherit from CDMTableBase "
        )

    # Explicitly mark as concrete
    cls.__abstract__ = False

    return cls



from datetime import datetime, date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Device_Exposure(Base):
    __tablename__ = 'device_exposure'
    # identifier
    device_exposure_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True)
    # temporal
    device_exposure_start_date: so.Mapped[date] = so.mapped_column(sa.Date)
    device_exposure_start_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    device_exposure_end_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    device_exposure_end_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    # strings
    unique_device_id: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    device_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(100))
    # numeric
    quantity: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer)
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='de_fk_1'))
    provider_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('provider.provider_id', name='de_fk_2'))
    visit_occurrence_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_occurrence.visit_occurrence_id', name='de_fk_3'))
    visit_detail_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_detail.visit_detail_id', name='de_fk_4'))
    # concept fks
    device_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='de_fk_5'))
    device_type_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='de_fk_6'))    
    device_source_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='de_fk_7'))
    # relationships
    person: so.Mapped['Person'] = so.relationship(foreign_keys=[person_id])
    provider: so.Mapped[Optional['Provider']] = so.relationship(foreign_keys=[provider_id])
    visit_occurrence: so.Mapped[Optional['Visit_Occurrence']] = so.relationship(foreign_keys=[visit_occurrence_id])
    visit_detail: so.Mapped[Optional['Visit_Detail']] = so.relationship(foreign_keys=[visit_detail_id])
    # concept relationships
    device_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[device_concept_id])
    device_type_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[device_type_concept_id])
    device_source_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[device_source_concept_id])
    


from datetime import datetime, date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so
from sqlalchemy.ext.hybrid import hybrid_property

from .modifiable_table import Modifiable_Table
from ...db import Base
from ...conventions.concept_enumerators import ModifierFields

class Procedure_Occurrence(Modifiable_Table):
    __tablename__ = 'procedure_occurrence'
    # identifier
    procedure_occurrence_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('modifiable_table.modifier_id', name='po_fk_1'), primary_key=True, autoincrement=True) 

    # temporal
    procedure_date: so.Mapped[date] = so.mapped_column(sa.Date)
    procedure_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime, nullable=True)
    # strings
    procedure_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    modifier_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    # numeric
    quantity: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer, nullable=True)
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='po_fk_2'))
    provider_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('provider.provider_id', name='po_fk_3'), nullable=True)
    visit_occurrence_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_occurrence.visit_occurrence_id', name='po_fk_4'), nullable=True)
    visit_detail_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_detail.visit_detail_id', name='po_fk_5'), nullable=True)
    # concept fks
    procedure_type_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='po_fk_6'))
    modifier_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='po_fk_7'), nullable=True)
    procedure_concept_id: so.Mapped[int] = so.mapped_column(sa.BigInteger, sa.ForeignKey('concept.concept_id', name='po_fk_8'))
    procedure_source_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.BigInteger, sa.ForeignKey('concept.concept_id', name='po_fk_9'), nullable=True)
    # relationships
    person: so.Mapped[Optional['Person']] = so.relationship(foreign_keys=[person_id])
    provider: so.Mapped[Optional['Provider']] = so.relationship(foreign_keys=[provider_id])
    visit_occurrence: so.Mapped[Optional['Visit_Occurrence']] = so.relationship(foreign_keys=[visit_occurrence_id])
    visit_detail: so.Mapped[Optional['Visit_Detail']] = so.relationship(foreign_keys=[visit_detail_id])
    # concept_relationships
    procedure_type_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[procedure_type_concept_id])
    modifier_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[modifier_concept_id])
    procedure_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[procedure_concept_id])
    procedure_source_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[procedure_source_concept_id])


    def __init__(self, 
                 person_id,
                 procedure_date,
                 procedure_type_concept_id,
                 procedure_concept_id,
                 procedure_datetime=None,
                 procedure_source_value=None,
                 modifier_source_value=None,
                 quantity=None,
                 provider_id=None,
                 visit_occurrence_id=None,
                 visit_detail_id=None,
                 modifier_concept_id=None,
                 procedure_source_concept_id=None,
                 *args, 
                 **kwargs):
        super().__init__(person_id=person_id,
                         procedure_date=procedure_date,
                         procedure_datetime=procedure_datetime,
                         procedure_source_value=procedure_source_value,
                         modifier_source_value=modifier_source_value,
                         quantity=quantity,
                         provider_id=provider_id,
                         visit_occurrence_id=visit_occurrence_id,
                         visit_detail_id=visit_detail_id,
                         procedure_type_concept_id=procedure_type_concept_id,
                         modifier_concept_id=modifier_concept_id,
                         procedure_concept_id=procedure_concept_id,
                         modifier_of_field_concept_id = ModifierFields.procedure_occurrence_id.value,
                         *args, **kwargs)


    __mapper_args__ = {
        "polymorphic_identity": "procedure",
        'inherit_condition': (procedure_occurrence_id == Modifiable_Table.modifier_id)
    }

    @hybrid_property
    def event_date(self):
        return  self.procedure_datetime.date() if self.procedure_datetime is not None else self.procedure_date

    @property
    def proc_label(self):
        if self.procedure_concept:
            return self.procedure_concept.concept_name
    


# tests/fixtures/load_athena.py

import pytest
from pathlib import Path
from omop_alchemy.cdm.base import bootstrap
import sqlalchemy as sa
from sqlalchemy.orm import sessionmaker

from omop_alchemy.cdm.model.vocabulary import (
    Domain,
    Vocabulary,
    Concept_Class,
    Relationship,
    Concept,
    Concept_Ancestor,
    Concept_Relationship,
)

ATHENA_LOAD_ORDER = [
    Domain,
    Vocabulary,
    Concept_Class,
    Relationship,
    Concept,
    Concept_Ancestor,
    Concept_Relationship,
]


# def load_athena_vocab(session, base_path: Path):
#     for model in ATHENA_LOAD_ORDER:
#         model.load_csv(
#             session,
#             base_path / f"{model.__tablename__}.csv",
#         )
#     session.flush()



@pytest.fixture(scope="session")
def engine():
    """
    In-memory SQLite database for tests.
    """
    engine = sa.create_engine(
        "sqlite+pysqlite:///:memory:",
        future=True,
    )

    bootstrap(engine, create=True)
    yield engine
    engine.dispose()


@pytest.fixture
def db_session(engine):
    """
    SQLAlchemy session per test.
    """
    Session = sessionmaker(bind=engine, future=True)
    session = Session()
    try:
        yield session
        session.commit()
    finally:
        session.rollback()
        session.close()


@pytest.fixture(scope="session")
def athena_vocab(engine):
    """
    Load a minimal, internally consistent Athena vocabulary
    using the real ORM CSV loader.
    """
    Session = sessionmaker(bind=engine, future=True)
    session = Session()

    base_path = (
        Path(__file__).parent
        / "fixtures"
        / "athena_source"
    )

    for model in ATHENA_LOAD_ORDER:
        csv_path = base_path / f"{model.__tablename__}.csv"
        if not csv_path.exists():
            raise RuntimeError(f"Missing vocab CSV: {csv_path}")

        model.load_csv(session, csv_path)

    session.commit()
    session.close()

    yield

def test_concept_loaded(db_session, athena_vocab):
    concept = db_session.get(Concept, 101)
    assert concept is not None
    assert concept.concept_name == "Lung cancer"
    assert concept.domain_id == "Condition"

def test_concept_ancestor(db_session, athena_vocab):
    ancestors = (
        db_session.query(Concept_Ancestor)
        .filter_by(descendant_concept_id=101)
        .all()
    )

    assert len(ancestors) == 1
    assert ancestors[0].ancestor_concept_id == 100

def test_condition_domain_is_correct(db_session, athena_vocab):
    condition = db_session.get(Concept, 101)
    assert condition.domain_id == "Condition"

    procedure = db_session.get(Concept, 201)
    assert procedure.domain_id == "Procedure"

def test_relationship_vocab_loaded(db_session, athena_vocab):
    rel = (
        db_session.query(Relationship)
        .filter_by(relationship_id="Uses drug")
        .one()
    )

    assert rel.reverse_relationship_id == "Drug used by"


from datetime import datetime, date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Survey_Conduct(Base):
    __tablename__ = 'survey_conduct'
    # identifier
    survey_conduct_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True) 
    # temporal
    survey_start_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    survey_start_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    survey_end_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    survey_end_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    # strings
    assisted_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    respondent_type_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(100))
    timing_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(100))
    collection_method_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(100))
    survey_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(100))
    survey_source_identifier: so.Mapped[Optional[str]] = so.mapped_column(sa.String(100))
    validated_survey_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(100))
    survey_version_number: so.Mapped[Optional[str]] = so.mapped_column(sa.String(20))
    # numeric
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='sc_fk_1'))
    provider_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('provider.provider_id', name='sc_fk_2'))
    visit_occurrence_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('visit_occurrence.visit_occurrence_id', name='sc_fk_3'))
    visit_detail_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('visit_detail.visit_detail_id', name='sc_fk_4'))    
    response_visit_occurrence_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('visit_occurrence.visit_occurrence_id', name='sc_fk_5'))
    # concept fks
    survey_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='sc_fk_6'))
    assisted_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='sc_fk_7'))
    respondent_type_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='sc_fk_8'))
    timing_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='sc_fk_9'))
    collection_method_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='sc_fk_10'))
    validated_survey_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='sc_fk_11'))
    survey_source_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='sc_fk_12'))

    # relationships
    person: so.Mapped[Optional['Person']] = so.relationship(foreign_keys=[person_id])
    provider: so.Mapped[Optional['Provider']] = so.relationship(foreign_keys=[provider_id])
    visit_occurrence: so.Mapped[Optional['Visit_Occurrence']] = so.relationship(foreign_keys=[visit_occurrence_id])
    visit_detail: so.Mapped[Optional['Visit_Detail']] = so.relationship(foreign_keys=[visit_detail_id])
    response_visit_occurrence: so.Mapped[Optional['Visit_Occurrence']] = so.relationship(foreign_keys=[response_visit_occurrence_id])

    # concept_relationships
    survey_concept_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[survey_concept_id])
    assisted_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[assisted_concept_id])
    respondent_type_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[respondent_type_concept_id])
    timing_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[timing_concept_id])
    collection_method_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[collection_method_concept_id])
    validated_survey_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[validated_survey_concept_id])
    survey_source_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[survey_source_concept_id])



import sqlalchemy as sa
import sqlalchemy.orm as so
from sqlalchemy.types import BigInteger
from datetime import date
from typing import Optional, List

from ...db import Base

class Concept(Base): 
    __tablename__ = 'concept'
    concept_id: so.Mapped[int] = so.mapped_column(sa.BigInteger, index=True, unique=True, primary_key=True)
    concept_name: so.Mapped[str] = so.mapped_column(sa.String(255))
    concept_code: so.Mapped[str] = so.mapped_column(sa.String(50))

    domain_id: so.Mapped[str] = so.mapped_column(sa.String(20), sa.ForeignKey('domain.domain_id', name='c_fk_1'))
    vocabulary_id: so.Mapped[str] = so.mapped_column(sa.String(20), sa.ForeignKey('vocabulary.vocabulary_id', name='c_fk_2'))
    concept_class_id: so.Mapped[str] = so.mapped_column(sa.String(20), sa.ForeignKey('concept_class.concept_class_id', name='c_fk_3'))
    
    domain: so.Mapped['Domain'] = so.relationship('Domain', primaryjoin='Concept.domain_id==Domain.domain_id', post_update=True, cascade="all")
    vocabulary: so.Mapped['Vocabulary'] = so.relationship('Vocabulary', primaryjoin='Concept.vocabulary_id==Vocabulary.vocabulary_id', post_update=True, cascade="all")#foreign_keys=[vocabulary_id], back_populates='vocabulary_concept')
    concept_class: so.Mapped['Concept_Class'] = so.relationship('Concept_Class', primaryjoin='Concept.concept_class_id==Concept_Class.concept_class_id', post_update=True, cascade="all")#foreign_keys=[concept_class_id])

    concept_relationships: so.Mapped[List['Concept_Relationship']] = so.relationship('Concept_Relationship', back_populates='concept_1', primaryjoin='Concept.concept_id==Concept_Relationship.concept_id_1')

    standard_concept: so.Mapped[Optional[str]]  = so.mapped_column(sa.String(1), nullable=True)
    valid_start_date: so.Mapped[date]  = so.mapped_column(sa.Date)
    valid_end_date: so.Mapped[date]  = so.mapped_column(sa.Date)
    invalid_reason: so.Mapped[Optional[str]]  = so.mapped_column(sa.String(1), nullable=True)

    def __repr__(self):
        return f'<Concept {self.concept_id} - {self.concept_code} ({self.concept_name})>'



import sqlalchemy as sa
import sqlalchemy.orm as so
from sqlalchemy.ext.hybrid import hybrid_property
from typing import List, Optional
from datetime import datetime
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy

from ...conventions import Modality, DiseaseEpisodeConcepts, TreatmentEpisode
from ...conventions.concept_enumerators import ModifierFields

from ...db import Base
from ..clinical.measurement import Measurement
from ..clinical.modifiable_table import Modifiable_Table
from ..concept_links import Concept_Links


class Episode(Modifiable_Table, Concept_Links):
    __tablename__ = 'episode'
    labels = {'episode': False, 'episode_object': False, 'episode_type': False, 'episode_source': False}

    # identifier
    episode_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('modifiable_table.modifier_id'), primary_key=True, autoincrement=True)

    # temporal
    episode_start_datetime: so.Mapped[datetime] = so.mapped_column(sa.DateTime)
    episode_end_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime, nullable=True)
    # strings
    episode_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    # numeric
    episode_number: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer, nullable=True)
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey("person.person_id"))
    episode_parent_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey("episode.episode_id"), nullable=True)
    # relationships
    person_object: so.Mapped['Person'] = so.relationship(back_populates="episodes", foreign_keys=[person_id])
    events: so.Mapped[List['Episode_Event']] = so.relationship(back_populates="episode_object", lazy="selectin")

    episode_parent_object: so.Mapped[Optional['Episode']] = so.relationship(foreign_keys=[episode_parent_id], remote_side=episode_id)
    children: so.Mapped[List['Episode']] = so.relationship(back_populates="episode_parent_object", lazy="selectin", foreign_keys=[episode_parent_id])

    __mapper_args__ = {
        "polymorphic_identity": "episode",
        'inherit_condition': (episode_id == Modifiable_Table.modifier_id)
    }

    def __init__(self, 
                 episode_concept_id=DiseaseEpisodeConcepts.episode_of_care.value, 
                 episode_type_concept_id=None,
                 episode_parent_id=None,
                 *args, 
                 **kwargs):
        # TODO: removing for now, as it is required for bulk loads if there's an arbitrary depth of parent/child
        # TODO: can this check be put somewhere else, made optional, supported thru limitation of depth instead?
        # if episode_concept_id != DiseaseEpisodeConcepts.episode_of_care.value:
        #     if episode_parent_id is None:
        #         raise ValueError('All episodes other than top level episode of care must have a parent value')
        super().__init__(episode_concept_id=episode_concept_id, 
                         episode_type_concept_id=episode_type_concept_id,
                         episode_parent_id=episode_parent_id, 
                         modifier_of_field_concept_id = ModifierFields.episode_id.value,
                         *args, **kwargs)
    

    def __repr__(self):
        ep_type = 'Treatment ' if self.is_tx else 'Diagnostic '
        return f'{ep_type} Episode: episode_id = {self.episode_id}'
    
    @property 
    def primary_ep(self):
        # todo: recursion risk - check depth on assignment?
        if self.is_overarching:
            return self.episode_id
        if self.episode_parent_object:
            return self.episode_parent_object.primary_ep

    @hybrid_property
    def is_overarching(self):
        return self.episode_concept_id == DiseaseEpisodeConcepts.episode_of_care.value

    @is_overarching.expression
    def is_overarching(cls):
        return cls.episode_concept_id == DiseaseEpisodeConcepts.episode_of_care.value
    
    @property
    def is_tx(self):
        return self.episode_concept_id in TreatmentEpisode.member_values()
        
    @property
    def is_dx(self):
        return self.episode_concept_id in DiseaseEpisodeConcepts.member_values()
    
    @property
    def event_date(self):
        return  self.episode_start_datetime.date()


    @hybrid_property
    def modality(self):
        if self.is_dx:
            raise ValueError('modality modifier not valid for diagnostic episodes')
        else:
            try:
                return [m for m in self.modifiers if m.measurement_concept_id in Modality.member_values()][0]
            except:
                return None

    @modality.inplace.setter
    def _modality_setter(self, value: Optional[Modality]) -> None:
        assert value in Modality    
        self.modifiers.append(Measurement(person_id = self.person_id,
                                          modified_object = self, 
                                          measurement_concept_id = value.value))





from datetime import datetime, date
from typing import Optional, List
import sqlalchemy as sa
import sqlalchemy.orm as so
from sqlalchemy.ext.hybrid import hybrid_property

from .modifiable_table import Modifiable_Table
from ..vocabulary.vocabulary import Concept
from ...db import Base
from ...conventions.concept_enumerators import ModifierFields

class Condition_Occurrence(Modifiable_Table):
    __tablename__ = 'condition_occurrence'
    validators = {}
    
    def __init__(self, 
                 person_id,
                 condition_start_date,
                 condition_concept_id,
                 condition_type_concept_id,
                 condition_start_datetime=None,
                 condition_end_date=None,
                 condition_end_datetime=None,
                 stop_reason=None,
                 condition_source_value=None,
                 condition_status_source_value=None,
                 provider_id=None,
                 visit_occurrence_id=None,
                 visit_detail_id=None,
                 condition_status_concept_id=None,
                 *args, 
                 **kwargs):
        condition_start_datetime = condition_start_datetime or datetime.combine(condition_start_date, datetime.min.time())
        super().__init__(person_id=person_id,
                         condition_start_date=condition_start_date,
                         condition_concept_id=condition_concept_id,
                         condition_type_concept_id=condition_type_concept_id,
                         condition_start_datetime=condition_start_datetime,
                         condition_end_date=condition_end_date,
                         condition_end_datetime=condition_end_datetime,
                         stop_reason=stop_reason,
                         condition_source_value=condition_source_value,
                         condition_status_source_value=condition_status_source_value,
                         provider_id=provider_id,
                         visit_occurrence_id=visit_occurrence_id,
                         visit_detail_id=visit_detail_id,
                         condition_status_concept_id=condition_status_concept_id,
                         modifier_of_field_concept_id = ModifierFields.condition_occurrence_id.value,
                         *args, **kwargs)


    # this is crazy slow we cannot operate like this - to investigate better way...
    # fine and very useful on individual insances - ridiculous on the bulk level

    # @so.reconstructor
    # def init_on_load(self):
    #     # TODO: need to create the class functions and the setter functions for this
    #     # Consider what happens if a new measurement is added - if this is fired on loading of
    #     # condition object, will addition of a new measurement be enough to trigger this update?
    #     # Probably not?
    #     self._p = False
    #     self._t, self._n, self._m, self._tnm = [], [], [], []
    #     for m in sorted(self.modifiers, key=lambda mod: mod.measurement_date):
    #         if not self._p and m.measurement_concept_id in self.validators['tnm'].path_stage_concepts:
    #             self._p=(m.measurement_date - self.condition_start_date).days < 30
    #         if m.measurement_concept_id in self.validators['tnm'].t_stage_concepts:
    #             self._t.append(m)
    #         elif m.measurement_concept_id in self.validators['tnm'].n_stage_concepts:
    #             self._n.append(m)
    #         elif m.measurement_concept_id in self.validators['tnm'].m_stage_concepts:
    #             self._m.append(m)
    #         elif m.measurement_concept_id in self.validators['tnm'].group_stage_concepts:
    #             self._tnm.append(m)

    # identifier
    condition_occurrence_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('modifiable_table.modifier_id'), primary_key=True, autoincrement=True)
    # temporal
    condition_start_date: so.Mapped[date] = so.mapped_column(sa.Date)
    condition_start_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime, nullable=True)
    condition_end_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date, nullable=True)
    condition_end_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime, nullable=True)
    # strings
    stop_reason: so.Mapped[Optional[str]] = so.mapped_column(sa.String(20), nullable=True)
    condition_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    condition_status_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='co_fk_1'))
    provider_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('provider.provider_id', name='co_fk_2'), nullable=True)
    visit_occurrence_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_occurrence.visit_occurrence_id', name='co_fk_3'), nullable=True)
    visit_detail_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_detail.visit_detail_id', name='co_fk_4'), nullable=True)
    # concept fks
    condition_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='co_fk_5'))
    condition_type_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='co_fk_6'))
    condition_status_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='co_fk_7'), nullable=True)
    condition_source_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='co_fk_8'), nullable=True)
    # relationships
    person: so.Mapped['Person'] = so.relationship(foreign_keys=[person_id])
    provider: so.Mapped[Optional['Provider']] = so.relationship(foreign_keys=[provider_id])
    visit_occurrence: so.Mapped[Optional['Visit_Occurrence']] = so.relationship(foreign_keys=[visit_occurrence_id])
    visit_detail: so.Mapped[Optional['Visit_Detail']] = so.relationship(foreign_keys=[visit_detail_id])
    # concept relationships
    condition_concept: so.Mapped['Concept'] = so.relationship(foreign_keys=[condition_concept_id])
    condition_type_concept: so.Mapped['Concept'] = so.relationship(foreign_keys=[condition_type_concept_id])
    condition_status_concept: so.Mapped['Concept'] = so.relationship(foreign_keys=[condition_status_concept_id])
    condition_source_concept: so.Mapped['Concept'] = so.relationship(foreign_keys=[condition_source_concept_id])

    __mapper_args__ = {
        "polymorphic_identity": "condition",
        'inherit_condition': (condition_occurrence_id == Modifiable_Table.modifier_id)
    }

    @classmethod
    def set_validators(cls):
        # putting this here so that we can defer the import until after the models have all been instantiated, otherwise 
        # it tries to query the concepts from non-existent tables - there may be a better way?
        from ...conventions.vocab_lookups import tnm_lookup
        cls.validators = {'tnm': tnm_lookup}

    @hybrid_property
    def condition_label(self):
        if self.condition_concept:
            return self.condition_concept.concept_name
    
    @hybrid_property
    def condition_code(self):
        if self.condition_concept:
            return self.condition_concept.concept_code
    
    @condition_code.expression
    @classmethod
    def condition_code(cls) -> sa.ColumnElement[Optional[str]]:
        return sa.select(Concept.concept_code).where(Concept.concept_id==cls.condition_concept_id).as_scalar()

    @condition_label.expression
    def _condition_label_expression(cls) -> sa.ColumnElement[Optional[str]]:
        return sa.cast("SQLColumnExpression[Optional[str]]", cls.condition_concept.concept_name)
    
    @property
    def event_date(self):
        return  self.condition_start_datetime.date() if self.condition_start_datetime is not None else self.condition_start_date


    # # TODO: Down the line we should consider if all of these oncology-extension-specific properties need to be refactored out into a subclass?

    # @hybrid_property
    # def path_confirmation(self):
    #     # returns true is any of the stage modifiers of this condition are of type pathological
    #     try:
    #         return self._p
    #     except:
    #         return False

    # @hybrid_property
    # def group_stage(self):
    #     # returns group stage modifiers of this condition, ordered by modifier date - at this point just returns earliest
    #     # TODO: do we want to accommodate most recent as well as first staging? 
    #     try:
    #         return self._tnm[0]
    #     except:
    #         return None

    # @hybrid_property
    # def t_stage(self):
    #     try:
    #         return self._t[0]
    #     except:
    #         return None
            
    # @hybrid_property
    # def n_stage(self):        
    #     try:
    #         return self._n[0]
    #     except:
    #         return None

    # @hybrid_property
    # def m_stage(self):
    #     try:
    #         return self._m[0]
    #     except:
    #         return None


import sqlalchemy as sa
import sqlalchemy.orm as so
from typing import Optional
from datetime import date
from omop_alchemy.cdm.base import (
    Base,
    cdm_table,
    CDMTableBase,
    optional_concept_fk,
)

@cdm_table
class Payer_Plan_Period(CDMTableBase, Base):
    __tablename__ = "payer_plan_period"

    payer_plan_period_id: so.Mapped[int] = so.mapped_column(primary_key=True)
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey("person.person_id"), nullable=False, index=True)
    payer_plan_period_start_date: so.Mapped[date] = so.mapped_column(nullable=False)
    payer_plan_period_end_date: so.Mapped[date] = so.mapped_column(nullable=False)

    payer_concept_id: so.Mapped[Optional[int]] = optional_concept_fk()
    payer_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    payer_source_concept_id: so.Mapped[Optional[int]] = optional_concept_fk()

    plan_concept_id: so.Mapped[Optional[int]] = optional_concept_fk()
    plan_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    plan_source_concept_id: so.Mapped[Optional[int]] = optional_concept_fk()

    sponsor_concept_id: so.Mapped[Optional[int]] = optional_concept_fk()
    sponsor_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    sponsor_source_concept_id: so.Mapped[Optional[int]] = optional_concept_fk()

    family_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)

    stop_reason_concept_id: so.Mapped[Optional[int]] = optional_concept_fk()
    stop_reason_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    stop_reason_source_concept_id: so.Mapped[Optional[int]] = optional_concept_fk()



from datetime import datetime, date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so
from decimal import Decimal
from sqlalchemy.ext.hybrid import hybrid_property

from ..concept_links import Concept_Links
from ...db import Base

class Observation(Base, Concept_Links):
    __tablename__ = 'observation'
    labels = {'observation': False, 'observation_type': False, 'observation_source': False, 'value_as': False, 'qualifier': False, 'unit': False, 'obs_event_field': False}

    # identifier
    observation_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True) 
    # temporal
    observation_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date, nullable=True)
    observation_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime, nullable=True)
    value_as_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime, nullable=True)
    # strings
    value_as_string: so.Mapped[Optional[str]] = so.mapped_column(sa.String(60), nullable=True)
    observation_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    unit_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    qualifier_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    # numeric
    value_as_number: so.Mapped[Optional[Decimal]] = so.mapped_column(sa.Numeric, nullable=True)
    # polymorphic fk
    observation_event_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('modifiable_table.modifier_id', name='o_fk_1'), nullable=True)
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='o_fk_2'))
    provider_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('provider.provider_id', name='o_fk_3'), nullable=True)
    visit_occurrence_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_occurrence.visit_occurrence_id', name='o_fk_4'), nullable=True)
    visit_detail_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_detail.visit_detail_id', name='o_fk_5'), nullable=True)
    # relationships
    person: so.Mapped['Person'] = so.relationship(foreign_keys=[person_id])
    provider: so.Mapped[Optional['Provider']] = so.relationship(foreign_keys=[provider_id])
    visit_occurrence: so.Mapped[Optional['Visit_Occurrence']] = so.relationship(foreign_keys=[visit_occurrence_id])
    visit_detail: so.Mapped[Optional['Visit_Detail']] = so.relationship(foreign_keys=[visit_detail_id])
    observed_object: so.Mapped[Optional['Modifiable_Table']] = so.relationship(foreign_keys=[observation_event_id])

    @property
    def observation_dt(self):
        # coalesce over observation datetime and observation_date returning date object or None
        return self.observation_datetime.date if self.observation_date else self.observation_date





from datetime import datetime, date
from typing import Optional, List
from sqlalchemy.ext.hybrid import hybrid_property
import sqlalchemy as sa
import sqlalchemy.orm as so

from ..concept_links import Concept_Links
from ...db import Base

class Person(Base):#, Concept_Links):
    __tablename__ = 'person'
    validators = {}
#    testing out @validates decorator instead and can't do this easily if these are added dynamically after all...
#    labels = {'gender': False, 'ethnicity': False, 'race': False, 'gender_source': False, 'ethnicity_source': False, 'race_source': False}

    # identifier
    person_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True)
    # temporal
    birth_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime, nullable=True)
    death_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime, nullable=True)
    # strings
    person_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    gender_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    race_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    ethnicity_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    # numeric
    year_of_birth: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer, nullable=True)
    month_of_birth: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer, nullable=True)
    day_of_birth: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer, nullable=True)
    # fks
    location_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('location.location_id', name='pr_fk_1'), nullable=True)
    provider_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('provider.provider_id', name='pr_fk_2'), nullable=True)
    care_site_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('care_site.care_site_id', name='pr_fk_3'), nullable=True)
    # concept fks
    gender_source_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='pr_fk_4'), nullable=True)
    ethnicity_source_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='pr_fk_5'), nullable=True)
    race_source_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='pr_fk_6'), nullable=True)
    gender_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='pr_fk_7'), default=0)
    ethnicity_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='pr_fk_8'), default=0)
    race_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='pr_fk_9'), default=0)
    # relationships
    location: so.Mapped[Optional['Location']] = so.relationship(foreign_keys=[location_id])
    provider: so.Mapped[Optional['Provider']] = so.relationship(foreign_keys=[provider_id])
    care_site: so.Mapped[Optional['Care_Site']] = so.relationship(foreign_keys=[care_site_id])
    death: so.Mapped[Optional['Death']] = so.relationship(back_populates="person", lazy="selectin")
    conditions: so.Mapped[List['Condition_Occurrence']] = so.relationship(back_populates="person", lazy="selectin")
    observations: so.Mapped[List['Observation']] = so.relationship(back_populates="person", lazy="selectin")
    episodes: so.Mapped[List['Episode']] = so.relationship(back_populates="person_object", lazy="selectin")
    procedures: so.Mapped[List['Procedure_Occurrence']] = so.relationship(back_populates="person", lazy="selectin")
    measurements: so.Mapped[List['Measurement']] = so.relationship(back_populates="person_object", lazy="selectin")
    # concept_relationships
    gender: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[gender_concept_id])
    ethnicity: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[ethnicity_concept_id])
    race: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[race_concept_id])
    gender_source: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[gender_source_concept_id])
    ethnicity_source: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[ethnicity_source_concept_id])
    race_source: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[race_source_concept_id])

    @classmethod
    def set_validators(cls):
        # putting this here so that we can defer the import until after the models have all been instantiated, otherwise 
        # it tries to query the concepts from non-existent tables - there may be a better way?
        from ...conventions.vocab_lookups import gender_lookup, ethnicity_lookup, race_lookup
        cls.validators = {'gender': gender_lookup, 'ethnicity': ethnicity_lookup, 'race': race_lookup}

    # don't love how repetitive this is, but haven't worked out how to add functions dynamically that
    # include decorators, so it'll do for now...
    @so.validates('gender_concept_id')
    def validate_gender(self, key, gender_concept_id):
        if gender_concept_id and gender_concept_id not in self.validators['gender']:
            raise ValueError("failed validation - gender concept")
        return gender_concept_id
    
    @so.validates('race_concept_id')
    def validate_race(self, key, race_concept_id):
        if race_concept_id and race_concept_id not in self.validators['race']:
            raise ValueError("failed validation - race concept")
        return race_concept_id
    
    @so.validates('ethnicity_concept_id')
    def validate_ethnicity(self, key, ethnicity_concept_id):
        if ethnicity_concept_id and ethnicity_concept_id not in self.validators['ethnicity']:
            raise ValueError("failed validation - ethnicity concept")
        return ethnicity_concept_id
    
    @so.validates('year_of_birth')
    def validate_yob(self, key, year_of_birth):
        if year_of_birth and (year_of_birth < 1900 or year_of_birth > datetime.now().year):
            raise ValueError("failed validation - year of birth out of range")
        return year_of_birth
        
    @so.validates('month_of_birth')
    def validate_mob(self, key, month_of_birth):
        if month_of_birth and (month_of_birth < 1 or month_of_birth > 12):
            raise ValueError("failed validation - month of birth out of range")
        return month_of_birth
        
    @so.validates('day_of_birth')
    def validate_day_of_birth(self, key, day_of_birth):
        # this is run before actually storing to attributes, so can't check for like 30th Feb using this form
        if day_of_birth and (day_of_birth < 1 or day_of_birth > 31):
            raise ValueError(f'failed validation - day of birth out of range')
        return day_of_birth

    def __repr__(self):
        age = self.age
        p = f'Person: person_id = {self.person_id}'
        if age != {}:
            y = age['age_years']
            d = age['age_days']
            p += f'; age={y} years and {d} days'
        # todo: this can actually be a bit annoying as it triggers a lazy load so depending on session scope, consider removing?
        if self.gender:
            try:
                p += f'; {self.gender.concept_name}'
            except:
                ...
        return p

    def age_calc(self, age_at):
        if self.dob is None:
            return {}
        age = (age_at - self.dob).days
        return {'age_total': age, 'age_years': age // 365, 'age_days': age % 365}

    @property
    def dob(self):
        if self.birth_datetime:
            return self.birth_datetime.date()
        if self.year_of_birth is None:
            return None
        day = self.day_of_birth or 1
        month = self.month_of_birth or 1
        return date(self.year_of_birth, month, day)

    @property
    def age(self, age_at=None):
        age_at = age_at or date.today()
        age_at = min(age_at, self.death_datetime.date()) if self.death_datetime is not None else age_at
        return self.age_calc(age_at)
    
    @hybrid_property
    def gender_label(self):
        if self.gender:
            return self.gender.concept_name

    @hybrid_property
    def primary_dx_eps(self):
        return [e for e in self.episodes if e.is_overarching]

    @hybrid_property
    def current_cond(self):
        if self.conditions:
            return sorted(self.conditions, key=lambda x: x.event_date)[-1]

    @hybrid_property
    def age_current_dx(self):
        cc = self.current_cond
        if cc:
            return self.age_calc(age_at = cc.event_date)


import sqlalchemy as sa
import sqlalchemy.orm as so
from sqlalchemy.ext.hybrid import hybrid_property
from typing import List, Optional
from datetime import datetime, date
from ...db import Base
from ...conventions.concept_enumerators import ModifierFields
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy


class Episode_Event(Base):
    __tablename__ = 'episode_event'
    # identifier
    episode_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey("episode.episode_id", name='ee_fk_2'), primary_key=True)
    event_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('modifiable_table.modifier_id', name='ee_fk_1'), primary_key=True)
    episode_event_field_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='ee_fk_3'), primary_key=True)
    
    episode_object: so.Mapped['Episode'] = so.relationship(foreign_keys=[episode_id])
    event_polymorphic: so.Mapped['Modifiable_Table'] = so.relationship(foreign_keys=[event_id])
    
    episode_start_datetime: AssociationProxy[datetime] = association_proxy("episode_object", "episode_start_datetime")

    @property
    def primary_ep(self):
        if self.episode_object:
            return self.episode_object.primary_ep


    def __init__(self, 
                 episode_id, 
                 event_id,
                 episode_event_field_concept_id):
        if episode_event_field_concept_id == ModifierFields.episode_id.value:
            raise ValueError('Object of type episode cannot be an episode event')
        super().__init__(episode_id=episode_id, 
                         event_id=event_id,
                         episode_event_field_concept_id=episode_event_field_concept_id)


"""
Update: moving these into the conventions.constructs submodule, as they make a lot more sense there, and can be split out into understandable chunks - getting too unwieldy here!
"""

# from datetime import date, datetime, time
# from typing import Optional, List
# from decimal import Decimal
# from itertools import chain
# import sqlalchemy as sa
# import sqlalchemy.orm as so
# from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
# from sqlalchemy.ext.hybrid import hybrid_property

# from ...db import Base
# from ...conventions.concept_enumerators import CancerProcedureTypes
# from .episode import Episode
# from .episode_event import Episode_Event
# from ..clinical.person import Person
# from ..clinical.modifiable_table import Modifiable_Table
# from ..clinical.condition_occurrence import Condition_Occurrence
# from ..clinical.drug_exposure import Drug_Exposure
# from ..clinical.procedure_occurrence import Procedure_Occurrence
# from ..clinical.observation import Observation
# from ..vocabulary.concept import Concept
# from ..vocabulary.concept_ancestor import Concept_Ancestor
# from ...conventions.concept_enumerators import ModifierFields, TreatmentEpisode, DiseaseEpisodeConcepts, DemographyConcepts


# # select all conditions that have been associated with an episode - TBC if we need to add filtering for overarching?

# diagnosis = so.aliased(Condition_Occurrence, flat=True)

# dx_episode_subquery = (
#     sa.select(
#         Episode_Event.episode_id.label('dx_episode_id'),
#         diagnosis.person_id,
#         diagnosis.condition_start_date,
#         Concept.concept_code
#     )
#     .join(
#         Episode_Event,
#         Episode_Event.event_polymorphic.of_type(diagnosis)
#     )
#     .join(
#         Concept,
#         Concept.concept_id==diagnosis.condition_concept_id
#     )
#     .subquery()
# )

# condition_ep_join = sa.join(Condition_Occurrence, Episode_Event, 
#                              sa.and_(Condition_Occurrence.condition_occurrence_id==Modifiable_Table.modifier_id,
#                                      Modifiable_Table.modifier_id==Episode_Event.event_id,
#                                      Episode_Event.episode_event_field_concept_id==ModifierFields.condition_occurrence_id.value))

# class Condition_Episode(Base):
#     __table__ = condition_ep_join
#     episode_id = Episode_Event.episode_id
#     person_id = Condition_Occurrence.person_id
#     condition_occurrence_id = Condition_Occurrence.condition_occurrence_id  
#     condition_concept_id = Condition_Occurrence.condition_concept_id
#     condition_code = Condition_Occurrence.condition_code
#     modifier_concepts = Condition_Occurrence.modifier_concepts
#     episode_start_datetime = Episode_Event.episode_start_datetime

#     #person_object: so.Mapped['Person_Episodes'] = so.relationship(back_populates="condition_episodes", foreign_keys=[person_id])


# # select all tx episodes that have at least one drug administration event 
# # and find the start of chemo administration for that episode
# # todo: should this filter on anti-cancer therapies?
# # no - this does not need to filter on anti-cancer therapies only, because we only
# # pull in drugs that have been attached to an episode

# systemic_therapy = so.aliased(Drug_Exposure, flat=True)
# radiation_therapy = so.aliased(Procedure_Occurrence, flat=True)

# systemic_therapy_subquery = (
#     sa.select(
#         Episode.person_id,
#         Episode.episode_id.label('sact_episode_id'),
#         Episode.episode_parent_id,
#         systemic_therapy.drug_exposure_start_date
#     )
#     .join(
#          Episode_Event,
#          Episode.episode_id==Episode_Event.episode_id
#      )
#     .join_from(
#         Episode_Event,
#         Episode_Event.event_polymorphic.of_type(systemic_therapy)
#     )
#     .subquery()
# )

# radiation_therapy_subquery = (
#     sa.select(
#         Episode.person_id,
#         Episode.episode_id.label('rad_episode_id'),
#         Episode.episode_parent_id,
#         radiation_therapy.procedure_datetime
#     )
#     .join(
#          Episode_Event,
#          Episode.episode_id==Episode_Event.episode_id
#      )
#     .join_from(
#         Episode_Event,
#         Episode_Event.event_polymorphic.of_type(radiation_therapy)
#     )
#     .subquery()
# )

# systemic_therapy_start = (
#     sa.select(
#         systemic_therapy_subquery.c.person_id,
#         systemic_therapy_subquery.c.sact_episode_id,
#         systemic_therapy_subquery.c.episode_parent_id,
#         sa.func.min(systemic_therapy_subquery.c.drug_exposure_start_date).label('sact_start')
#     )
#     .group_by(        
#         systemic_therapy_subquery.c.person_id,        
#         systemic_therapy_subquery.c.episode_parent_id,
#         systemic_therapy_subquery.c.sact_episode_id
#     )
#     .subquery()
# )

# radiation_therapy_start = (
#     sa.select(
#         radiation_therapy_subquery.c.person_id,
#         radiation_therapy_subquery.c.rad_episode_id,
#         radiation_therapy_subquery.c.episode_parent_id,
#         sa.func.min(radiation_therapy_subquery.c.procedure_datetime).label('rt_start')
#     )
#     .group_by(        
#         radiation_therapy_subquery.c.person_id,        
#         radiation_therapy_subquery.c.episode_parent_id,
#         radiation_therapy_subquery.c.rad_episode_id
#     )
#     .subquery()
# )

# diagnosis = so.aliased(Condition_Occurrence, flat=True)

# dx_subquery = (
#     sa.select(
#         Episode.person_id,
#         Episode.episode_id.label('dx_episode_id'),
#         diagnosis.condition_start_date
#     )
#     .join(
#          Episode_Event,
#          Episode.episode_id==Episode_Event.episode_id
#      )
#     .join_from(
#         Episode_Event,
#         Episode_Event.event_polymorphic.of_type(diagnosis)
#     )
#     .subquery()
# )

# systemic_therapy_with_dx = (
#     sa.join(
#         dx_subquery, 
#         systemic_therapy_start, 
#         sa.and_(
#             systemic_therapy_start.c.episode_parent_id==dx_subquery.c.dx_episode_id, 
#             systemic_therapy_start.c.person_id==dx_subquery.c.person_id
#         ),
#         isouter=True
#     )
# )

# rt_therapy_with_dx = (
#     sa.join(
#         dx_subquery, 
#         radiation_therapy_start, 
#         sa.and_(
#             radiation_therapy_start.c.episode_parent_id==dx_subquery.c.dx_episode_id, 
#             radiation_therapy_start.c.person_id==dx_subquery.c.person_id
#         ),
#         isouter=True
#     )
# )

# class Systemic_Therapy_Episode(Base):
#     __table__ = systemic_therapy_with_dx
#     episode_id = systemic_therapy_start.c.sact_episode_id
#     person_id = systemic_therapy_start.c.person_id
#     sact_start = so.column_property(systemic_therapy_start.c.sact_start)
    
#     dx_ep_id = dx_subquery.c.dx_episode_id

#     dx_object: so.Mapped[Optional['Episode']] = so.relationship(foreign_keys=[dx_ep_id], viewonly=True)
#     #person_object: so.Mapped['Person_Episodes'] = so.relationship(back_populates="sact_episodes", foreign_keys=[person_id])
#     episode_object: so.Mapped['Episode'] = so.relationship(foreign_keys=[episode_id])
#     sact_events: AssociationProxy[List['Episode_Event']] = association_proxy("episode_object", "events")

#     @property
#     def episode_agents(self):
#         return list(set([s.event_polymorphic.drug_label for s in self.sact_events if s.event_polymorphic.polymorphic_label=='drug_exposure']))

# class Radiation_Therapy_Episode(Base):
#     __table__ = rt_therapy_with_dx
#     episode_id = radiation_therapy_start.c.rad_episode_id
#     person_id = radiation_therapy_start.c.person_id
#     rt_start = so.column_property(radiation_therapy_start.c.rt_start)
#     dx_ep_id = dx_subquery.c.dx_episode_id

#     dx_object: so.Mapped[Optional['Episode']] = so.relationship(foreign_keys=[dx_ep_id], viewonly=True)
#     #person_object: so.Mapped['Person_Episodes'] = so.relationship(back_populates="sact_episodes", foreign_keys=[person_id])
#     episode_object: so.Mapped['Episode'] = so.relationship(foreign_keys=[episode_id])
#     rt_events: AssociationProxy[List['Episode_Event']] = association_proxy("episode_object", "events")


# class Person_Episodes(Person):
#     #condition_episodes: so.Mapped[List['Condition_Episode']] = so.relationship(back_populates="person_object", lazy="selectin")
#     sact_episodes: so.Mapped[List['Systemic_Therapy_Episode']] = so.relationship(back_populates="person_object", order_by='Systemic_Therapy_Episode.sact_start')

#     @property
#     def all_agents(self):
#         return list(set(chain.from_iterable([se.episode_agents for se in self.sact_episodes])))


# Systemic_Therapy_Episode.person_object = so.relationship(
#     Person_Episodes,
#     primaryjoin=so.foreign(Systemic_Therapy_Episode.person_id) == Person_Episodes.person_id
# )

# class SACT_Event(Episode_Event):
#     __table__ = (
#         sa.select(Episode_Event)
#         .where(Episode_Event.episode_event_field_concept_id==ModifierFields.drug_exposure_id.value)
#         .subquery()
#     )
    
#     regimen_id = so.column_property(__table__.c.episode_id)
#     sact_id = so.column_property(__table__.c.event_id)
    
#     __mapper_args__ = {
#         'polymorphic_identity': 'sact_event',
#         'inherit_condition': (Episode_Event.event_id == __table__.c.event_id)
#     }

# class RT_Event(Episode_Event):
#     __table__ = (
#         sa.select(Episode_Event)
#         .where(Episode_Event.event_polymorphic.of_type(radiation_therapy))
#         .subquery()
#     )
    
#     regimen_id = so.column_property(__table__.c.episode_id)
#     rt_id = so.column_property(__table__.c.event_id)
    
#     __mapper_args__ = {
#         'polymorphic_identity': 'rt_event',
#         'inherit_condition': (Episode_Event.event_id == __table__.c.event_id)
#     }

# regimen_subquery = (
#     sa.select(Episode)
#     .where(Episode.episode_concept_id==TreatmentEpisode.treatment_regimen.value)
#     .subquery()
# )

# diagnosis_subquery = (
#     sa.select(Episode)
#     .where(Episode.episode_concept_id==DiseaseEpisodeConcepts.episode_of_care.value)
#     .subquery()
# )

# Regimen = so.with_polymorphic(Episode, [], selectable=regimen_subquery)
# Diagnosis = so.with_polymorphic(Episode, [], selectable=diagnosis_subquery)

# dx_with_regimen = (
#     sa.select(
#         Diagnosis.person_id,
#         Diagnosis.episode_id.label('dx_id'), 
#         Diagnosis.episode_start_datetime.label('dx_date'), 
#         sa.func.min(Regimen.episode_start_datetime).label('treatment_start'),
#         sa.func.max(Regimen.episode_end_datetime).label('treatment_end')
#     )
#     .join(Regimen, Regimen.episode_parent_id==Diagnosis.episode_id)
#     .group_by(Diagnosis.person_id, Diagnosis.episode_id, Diagnosis.episode_start_datetime)
#     .subquery()
# )

# dx_with_sact = (
#     sa.select(
#         Diagnosis.person_id,
#         Diagnosis.episode_id.label('dx_id'), 
#         Diagnosis.episode_start_datetime.label('dx_date'), 
#         sa.func.min(Drug_Exposure.drug_exposure_start_date).label('sact_start'),
#         sa.func.max(Drug_Exposure.drug_exposure_start_date).label('sact_end')
#     )
#     .join(Regimen, Regimen.episode_parent_id==Diagnosis.episode_id)
#     .join(SACT_Event, SACT_Event.regimen_id==Regimen.episode_id)
#     .join(Drug_Exposure, Drug_Exposure.drug_exposure_id==SACT_Event.sact_id)
#     .group_by(Diagnosis.person_id, Diagnosis.episode_id, Diagnosis.episode_start_datetime)
#     .subquery()
# )

# dx_with_rt = (
#     sa.select(
#         Diagnosis.person_id, 
#         Diagnosis.episode_id.label('dx_id'), 
#         Diagnosis.episode_start_datetime.label('dx_date'), 
#         sa.func.min(Procedure_Occurrence.procedure_date).label('rt_start'),
#         sa.func.max(Procedure_Occurrence.procedure_date).label('rt_end')
#     )
#     .join(Regimen, Regimen.episode_parent_id==Diagnosis.episode_id)
#     .join(RT_Event, RT_Event.regimen_id==Regimen.episode_id)
#     .join(Procedure_Occurrence, Procedure_Occurrence.procedure_occurrence_id==RT_Event.rt_id)
#     .group_by(Diagnosis.person_id, Diagnosis.episode_id, Diagnosis.episode_start_datetime)
#     .subquery()
# )

# class Dx_Treat_Start(Base):
#     __table__ = dx_with_regimen
#     person_id = dx_with_regimen.c.person_id
#     dx_id = dx_with_regimen.c.dx_id
#     dx_date = dx_with_regimen.c.dx_date
#     treatment_start = so.column_property(dx_with_regimen.c.treatment_start)
#     treatment_end = so.column_property(dx_with_regimen.c.treatment_end)

# class Dx_SACT_Start(Base):
#     __table__ = dx_with_sact
#     person_id = dx_with_sact.c.person_id
#     dx_id = dx_with_sact.c.dx_id
#     dx_date = dx_with_sact.c.dx_date
#     sact_start = so.column_property(dx_with_sact.c.sact_start)
#     sact_end = so.column_property(dx_with_sact.c.sact_end)

# class Dx_RT_Start(Base):
#     __table__ = dx_with_rt
#     person_id = dx_with_rt.c.person_id
#     dx_id = dx_with_rt.c.dx_id
#     dx_date = dx_with_rt.c.dx_date
#     rt_start = so.column_property(dx_with_rt.c.rt_start)
#     rt_end = so.column_property(dx_with_rt.c.rt_end)

# rth_ca = so.aliased(Concept_Ancestor, name='rth_ca')
# srg_ca = so.aliased(Concept_Ancestor, name='srg_ca')

# surgical = (
#     sa.select(
#         Concept.concept_name,
#         Concept.concept_code,
#         Concept.concept_id,
#         srg_ca.descendant_concept_id
#     )
#     .join(srg_ca, Concept.concept_id == srg_ca.descendant_concept_id)
#     .filter(srg_ca.ancestor_concept_id==CancerProcedureTypes.surgical_procedure.value)
#     .subquery()
# )

# historical_procedure = so.aliased(
#     Observation,
#     sa.select(Observation).where(
#         Observation.observation_concept_id == CancerProcedureTypes.historical_procedure.value
#     ).subquery(), 
#     'historical_procedure'
# )

# historical_surgery = (
#     sa.select(
#         historical_procedure.person_id,
#         historical_procedure.observation_id,
#         historical_procedure.value_as_concept_id,
#         historical_procedure.observation_datetime
#     ).join(
#         surgical, surgical.c.concept_id == historical_procedure.value_as_concept_id
#     ).subquery()
# )

# radiotherapy = (
#     sa.select(
#         rth_ca.descendant_concept_id.label('rt_id')
#     )
#     .select_from(rth_ca)
#     .filter(
#         sa.or_(
#             rth_ca.ancestor_concept_id==CancerProcedureTypes.rt_procedure.value,
#             rth_ca.ancestor_concept_id==CancerProcedureTypes.rn_procedure.value
#         )
#     )
#     .subquery()
# )
    
# surg_only = (
#     sa.join(
#         surgical,
#         radiotherapy,
#         radiotherapy.c.rt_id == surgical.c.descendant_concept_id,
#         isouter=True
#     )
# )

# surgical_procedure = (
#     sa.select(
#         Procedure_Occurrence.person_id,
#         Procedure_Occurrence.procedure_occurrence_id,
#         Procedure_Occurrence.procedure_concept_id,
#         Procedure_Occurrence.procedure_datetime,
#         surgical.c.concept_name,
#         surgical.c.concept_code
#     )
#     .join(surg_only, surgical.c.concept_id == Procedure_Occurrence.procedure_concept_id)
#     .filter(radiotherapy.c.rt_id == None)
#     .subquery()
# )

# class Dated_Surgical_Procedure(Base):
#     __table__ = surgical_procedure
#     procedure_occurrence_id = surgical_procedure.c.procedure_occurrence_id
#     person_id = surgical_procedure.c.person_id
#     procedure_concept_id = surgical_procedure.c.procedure_concept_id
#     concept_name = surgical_procedure.c.concept_name
#     concept_code = surgical_procedure.c.concept_code
#     procedure_datetime = so.column_property(surgical_procedure.c.procedure_datetime)


# class Historical_Surgical_Procedure(Base):
#     __table__ = historical_surgery
#     observation_id = historical_surgery.c.observation_id
#     person_id = historical_surgery.c.person_id
#     procedure_concept_id = historical_surgery.c.value_as_concept_id
#     history_datettime = so.column_property(historical_surgery.c.observation_datetime)





import sqlalchemy as sa
import sqlalchemy.orm as so
from typing import Optional
from datetime import date

from omop_alchemy.cdm.base import (
    Base,
    cdm_table,
    CDMTableBase,
    required_int,
    optional_int,
    required_concept_fk,
    optional_concept_fk,
)

@cdm_table
class Visit_Occurrence(CDMTableBase, Base):
    __tablename__ = "visit_occurrence"

    visit_occurrence_id: so.Mapped[int] = so.mapped_column(primary_key=True)
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey("person.person_id"),nullable=False,index=True)
    
    visit_concept_id: so.Mapped[int] = required_concept_fk(index=True)
    visit_start_date: so.Mapped[date] = so.mapped_column(sa.Date, nullable=False)
    visit_start_datetime: so.Mapped[Optional[date]] = so.mapped_column(sa.DateTime, nullable=True)
    visit_end_date: so.Mapped[date] = so.mapped_column(sa.Date, nullable=False)
    visit_end_datetime: so.Mapped[Optional[date]] = so.mapped_column(sa.DateTime, nullable=True)
    visit_type_concept_id: so.Mapped[int] = required_concept_fk(index=True)

    provider_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey("provider.provider_id"),nullable=True,index=True)
    care_site_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey("care_site.care_site_id"),nullable=True,index=True)

    visit_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    visit_source_concept_id: so.Mapped[Optional[int]] = optional_concept_fk(index=True)
    admitted_from_concept_id: so.Mapped[Optional[int]] = optional_concept_fk(index=True)
    admitted_from_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    discharged_to_concept_id: so.Mapped[Optional[int]] = optional_concept_fk(index=True)
    discharged_to_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)

    preceding_visit_occurrence_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey("visit_occurrence.visit_occurrence_id"),nullable=True,index=True)

    def __repr__(self) -> str:
        return f"<VisitOccurrence {self.visit_occurrence_id}>"



from .logging import get_logger, configure_logging
from .errors import CDMValidationError
from .type_management import perform_cast

__all__ = ["get_logger", "configure_logging", "CDMValidationError", "perform_cast"]


from datetime import date, datetime, time
from typing import Optional
from decimal import Decimal
import sqlalchemy as sa
import sqlalchemy.orm as so
from sqlalchemy.ext.hybrid import hybrid_property

from ...db import Base
from ..concept_links import Concept_Links

class Measurement(Base, Concept_Links):
    __tablename__ = 'measurement'
    labels = {'value_as': False, 'unit': False, 'measurement': False, 'measurement_source': False, 'measurement_type': False, 'operator': False, 'modifier_of_field': False}

    # identifier
    measurement_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True)
    # temporal
    measurement_date: so.Mapped[date] = so.mapped_column(sa.Date)
    measurement_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime, nullable=True)
    measurement_time: so.Mapped[Optional[time]] = so.mapped_column(sa.Time, nullable=True)
    # strings
    measurement_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    unit_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    value_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50), nullable=True)
    # numeric
    value_as_number: so.Mapped[Optional[Decimal]] = so.mapped_column(sa.Numeric, nullable=True)
    range_low: so.Mapped[Optional[Decimal]] = so.mapped_column(sa.Numeric, nullable=True)
    range_high: so.Mapped[Optional[Decimal]] = so.mapped_column(sa.Numeric, nullable=True)
    # polymorphic fk
    modifier_of_event_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('modifiable_table.modifier_id', name='m_fk_1'), nullable=True)
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='m_fk_2'))
    provider_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('provider.provider_id', name='m_fk_3'), nullable=True)
    visit_occurrence_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_occurrence.visit_occurrence_id', name='m_fk_4'), nullable=True)
    visit_detail_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_detail.visit_detail_id', name='m_fk_5'), nullable=True)
    # relationships
    person_object: so.Mapped['Person'] = so.relationship(foreign_keys=[person_id])
    provider_object: so.Mapped[Optional['Provider']] = so.relationship(foreign_keys=[provider_id])
    visit_occurrence_object: so.Mapped[Optional['Visit_Occurrence']] = so.relationship(foreign_keys=[visit_occurrence_id])
    visit_detail_object: so.Mapped[Optional['Visit_Detail']] = so.relationship(foreign_keys=[visit_detail_id])
    modified_object: so.Mapped[Optional['Modifiable_Table']] = so.relationship(foreign_keys=[modifier_of_event_id])


    @hybrid_property
    def measurement_label(self):
        if self.measurement_concept:
            return self.measurement_concept.concept_code


from sqlalchemy.orm import DeclarativeBase
from sqlalchemy import MetaData
from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlalchemy import text
from sqlalchemy.exc import IntegrityError

from omop_alchemy.cdm.utils import get_logger
from .typing import ORMTable

logger = get_logger(__name__)

NAMING_CONVENTIONS = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}

metadata = MetaData(naming_convention=NAMING_CONVENTIONS)

@event.listens_for(Engine, "connect")
def enable_sqlite_foreign_keys(dbapi_connection, connection_record):
    # SQLite only
    if dbapi_connection.__class__.__module__.startswith("sqlite3"):
        logger.debug("Enabling SQLite foreign key enforcement (PRAGMA foreign_keys=ON)")
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA defer_foreign_keys = ON;")
        cursor.close()


def explain_sqlite_fk_error(session, exc: IntegrityError, raise_error: bool = True) -> None:
    engine = session.get_bind()

    if engine.dialect.name != "sqlite":
        raise exc
    
    with engine.connect() as conn:
        logger.info("SQLite integrity violation detected, running PRAGMA foreign_key_check")
        rows = conn.execute(text("PRAGMA foreign_key_check")).fetchall()
        if len(rows) == 0:
            logger.info("No foreign key violations found by PRAGMA foreign_key_check")
            return
        logger.error(
            "SQLite foreign key violations found: %d issue(s)", 
            len(rows)
        )

    for r in rows:
        logger.debug(
            "FK violation: table=%s rowid=%s references=%s fk_index=%s",
            r[0], r[1], r[2], r[3]
        )
    if raise_error:
        raise exc


class Base(DeclarativeBase):
    metadata = metadata

def create_db(Base, engine):
    logger.debug("Database dialect: %s", engine.dialect.name)
    Base.metadata.create_all(engine)

def bootstrap(engine, *, create: bool = True):
    """
    Initialise OMOP schema against a provided SQLAlchemy engine.
    """
    from omop_alchemy.cdm.model import __all__ as model_modules
    # for module_name in model_modules:
    #     __import__(f"omop_alchemy.model.{module_name}")
    #     print(module_name)
    logger.info("Bootstrapping OMOP schema (create=%s)", create)
    if create:
        logger.info("Schema creation enabled")
        create_db(Base, engine)
    else:
        logger.info("Schema creation skipped (existing schema assumed)")

        
def get_table_by_name(tablename: str) -> ORMTable | None:
    for cls in Base.__subclasses__():
        if cls.__tablename__ == tablename.lower().strip():
            if isinstance(cls, ORMTable):
                return cls
    return None



# from .care_site import Care_Site
# from .location_history import Location_History
# from .location import Location
# from .provider import Provider

# __all__ = [Care_Site, Location_History, Location, Provider]


import sqlalchemy as sa
import sqlalchemy.orm as so
from typing import Optional, TYPE_CHECKING
from datetime import date, datetime

from omop_alchemy.cdm.base import (
    Base,
    PersonScoped, 
    ReferenceContextMixin,
    CDMTableBase,
    cdm_table, 
    required_concept_fk,
    optional_concept_fk,
    optional_int,
    DomainValidationMixin,
    ExpectedDomain,
    HealthSystemContext
)

if TYPE_CHECKING:
    from ..vocabulary import Concept
    from ..clinical import Visit_Occurrence, Visit_Detail, Person
    from ..health_system import Provider

@cdm_table

class Note_NLP(CDMTableBase, Base):
    __tablename__ = "note_nlp"

    note_nlp_id: so.Mapped[int] = so.mapped_column(primary_key=True)

    note_id: so.Mapped[int] = so.mapped_column(
        sa.ForeignKey("note.note_id"),
        nullable=False,
        index=True,
    )

    section_concept_id: so.Mapped[Optional[int]] = optional_concept_fk()

    snippet: so.Mapped[Optional[str]] = so.mapped_column(
        sa.String(250),
        nullable=True,
    )

    offset: so.Mapped[Optional[str]] = so.mapped_column(
        sa.String(50),
        nullable=True,
    )

    lexical_variant: so.Mapped[str] = so.mapped_column(
        sa.String(250),
        nullable=False,
    )

    note_nlp_concept_id: so.Mapped[Optional[int]] = optional_concept_fk()
    note_nlp_source_concept_id: so.Mapped[Optional[int]] = optional_concept_fk()

    nlp_system: so.Mapped[Optional[str]] = so.mapped_column(
        sa.String(250),
        nullable=True,
    )

    nlp_date: so.Mapped[date] = so.mapped_column(sa.Date, nullable=False)
    nlp_datetime: so.Mapped[Optional[date]] = so.mapped_column(
        sa.DateTime,
        nullable=True,
    )

    term_exists: so.Mapped[Optional[str]] = so.mapped_column(
        sa.String(1),
        nullable=True,
    )

    term_temporal: so.Mapped[Optional[str]] = so.mapped_column(
        sa.String(50),
        nullable=True,
    )

    term_modifiers: so.Mapped[Optional[str]] = so.mapped_column(
        sa.String(2000),
        nullable=True,
    )

    def __repr__(self) -> str:
        return f"<NoteNLP {self.note_nlp_id}>"


from .concept_ancestor import Concept_Ancestor
from .concept_class import Concept_Class
from .concept_relationship import Concept_Relationship
from .concept_synonym import Concept_Synonym
from .concept import Concept
from .domain import Domain
from .drug_strength import Drug_Strength
from .relationship import Relationship
from .source_to_concept_map import Source_To_Concept_Map
from .vocabulary import Vocabulary

__all__ = [Concept_Ancestor, Concept_Class, Concept_Relationship, 
           Concept_Synonym, Concept, Domain, Drug_Strength, Relationship, 
           Source_To_Concept_Map, Vocabulary]


from datetime import datetime, date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so
from decimal import Decimal

from ...db import Base

class Specimen(Base):
    __tablename__ = 'specimen'
    # identifier
    specimen_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True) 
    # temporal
    specimen_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    specimen_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    # strings
    specimen_source_id: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    specimen_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    unit_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    anatomic_site_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    disease_status_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    # numeric    
    quantity: so.Mapped[Optional[Decimal]] = so.mapped_column(sa.Numeric)
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='s_fk_1'))
    # concept fks
    specimen_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='s_fk_2'))
    specimen_type_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='s_fk_3'))
    unit_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='s_fk_4'))
    anatomic_site_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='s_fk_5'))
    disease_status_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='s_fk_6'))
    # relationships
    person: so.Mapped[Optional['Person']] = so.relationship(foreign_keys=[person_id])
    # concept_relationships
    specimen_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[specimen_concept_id])
    specimen_type_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[specimen_type_concept_id])
    unit_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[unit_concept_id])
    anatomic_site_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[anatomic_site_concept_id])
    disease_status_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[disease_status_concept_id])


from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so
from decimal import Decimal
from ...db import Base

class Location(Base):
    __tablename__ = 'location'
    # identifier
    location_id: so.Mapped[int] = so.mapped_column(primary_key=True, unique=True, autoincrement=True) 
    # temporal
    # strings
    address_1: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    address_2: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    city: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    # todo: can we make this string length configurable? srsly states aren't always 2 chars guys
    state: so.Mapped[Optional[str]] = so.mapped_column(sa.String(2))
    zip_code: so.Mapped[Optional[str]] = so.mapped_column(sa.String(9)) # note column name change for python compatibility
    county: so.Mapped[Optional[str]] = so.mapped_column(sa.String(20))
    country: so.Mapped[Optional[str]] = so.mapped_column(sa.String(100))
    location_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    # numeric
    latitude: so.Mapped[Optional[Decimal]] = so.mapped_column(sa.Numeric)
    longitude: so.Mapped[Optional[Decimal]] = so.mapped_column(sa.Numeric)
    # fks
    # concept fks
    # relationships
    # concept relationships




from datetime import date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Location_History(Base):
    __tablename__ = 'location_history'
    # identifier
    location_history_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True) 
    # temporal
    start_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    end_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    # strings
    # numeric
    entity_id: so.Mapped[Optional[int]] = so.mapped_column(sa.Integer)
    # fks
    location_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('location.location_id', name='lh_fk_1'))
    domain_id: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    # concept fks
    relationship_type_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='lh_fk_2'))
    # relationships
    location: so.Mapped[Optional['Location']] = so.relationship(foreign_keys=[location_id])
    # concept relationships
    relationship_type_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[relationship_type_concept_id])


import pytest
# from omop_alchemy.db import Config

# def test_config_create():
#     assert Config() is not None

# def test_config_contents():
# # TODO: figure out how to override config file to make sure we can reflect settings correctly
#     ...

# def test_schema_creation():
#     ...


# @pytest.fixture
# def db():
# #    create sqlite temp db that can be used to run tests
#    engine = create_engine(
#        "sqlite:///:memory:", connect_args={"check_same_thread": False}
#    )
#    TestSession = sessionmaker(autocommit=False, autoflush=False, bind=engine)
#    Base.metadata.create_all(bind=engine)
#    return TestSession()


import sqlalchemy as sa
import sqlalchemy.orm as so
from ...db import Base

class Concept_Class(Base): 
    __tablename__ = 'concept_class'
    concept_class_id: so.Mapped[str] = so.mapped_column(sa.String(20), index=True, unique=True, primary_key=True)
    concept_class_name: so.Mapped[str] = so.mapped_column(sa.String(255))
    concept_class_concept_id: so.Mapped[int] = so.mapped_column(sa.BigInteger, sa.ForeignKey('concept.concept_id', name='cc_fk_1'))

#    concept: so.Mapped['Concept'] = so.relationship('Concept', primaryjoin='Concept.concept_id==Concept_Class.concept_class_concept_id', post_update=True)

   # concept_class_concept: so.Mapped['Concept'] = so.relationship(foreign_keys=[concept_class_concept_id])
    concept_class_concept: so.Mapped['Concept'] = so.relationship('Concept', primaryjoin='Concept_Class.concept_class_concept_id==Concept.concept_id')

    def __repr__(self):
        return f'<ConceptClass {self.concept_class_id} - {self.concept_class_name}>'



import os
from dotenv import load_dotenv
from pathlib import Path
from .cdm.utils import get_logger

ROOT_PATH = Path(__file__).parent
TEST_PATH = Path(__file__).parent.parent / "tests"

logger = get_logger(__name__)

def load_environment(dotenv: bool = True) -> None:
    """
    Explicitly load environment variables for the application.
    Safe: does not log sensitive values.
    """
    if dotenv and load_dotenv():
        logger.info("Environment variables loaded from .env file")
    else:
        logger.debug("No .env file loaded")

def get_engine_name() -> str:
    engine = os.getenv("ENGINE", "")
    if engine:
        logger.info("Database engine configured")
    else:
        logger.warning("Database engine not configured")
    return engine


from datetime import datetime, date
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Visit_Detail(Base):
    __tablename__ = 'visit_detail'
    # identifier
    visit_detail_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True) 
    # temporal
    visit_detail_start_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    visit_detail_start_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    visit_detail_end_date: so.Mapped[Optional[date]] = so.mapped_column(sa.Date)
    visit_detail_end_datetime: so.Mapped[Optional[datetime]] = so.mapped_column(sa.DateTime)
    # strings
    admitted_from_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    visit_detail_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    discharge_to_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    # numeric
    # fks
    person_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('person.person_id', name='vd_fk_1'))
    provider_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('provider.provider_id', name='vd_fk_2'), nullable=True)
    visit_occurrence_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('visit_occurrence.visit_occurrence_id', name='vd_fk_3'))
    visit_detail_parent_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('visit_detail.visit_detail_id', name='vd_fk_4'), nullable=True)
    preceding_visit_detail_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('visit_detail.visit_detail_id', name='vd_fk_5'), nullable=True)
    care_site_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('care_site.care_site_id', name='vd_fk_6'), nullable=True)
    # concept fks
    visit_detail_type_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='vd_fk_7'))
    discharge_to_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='vd_fk_8'), nullable=True)
    visit_detail_source_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='vd_fk_9'), nullable=True)
    visit_detail_concept_id: so.Mapped[int] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='vd_fk_10'))
    admitted_from_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='vd_fk_11'), nullable=True)

    # relationships
    person: so.Mapped[Optional['Person']] = so.relationship(foreign_keys=[person_id])
    provider: so.Mapped[Optional['Provider']] = so.relationship(foreign_keys=[provider_id])
    visit_occurrence: so.Mapped[Optional['Visit_Occurrence']] = so.relationship(foreign_keys=[visit_occurrence_id])
    visit_detail_parent: so.Mapped[Optional['Visit_Detail']] = so.relationship(foreign_keys=[visit_detail_parent_id])
    preceding_visit_detail: so.Mapped[Optional['Visit_Detail']] = so.relationship(foreign_keys=[preceding_visit_detail_id])
    care_site: so.Mapped[Optional['Care_Site']] = so.relationship(foreign_keys=[care_site_id])

    # concept_relationships
    visit_detail_type_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[visit_detail_type_concept_id])
    discharge_to_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[discharge_to_concept_id])
    visit_detail_source_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[visit_detail_source_concept_id])
    visit_detail_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[visit_detail_concept_id])
    admitted_from_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[admitted_from_concept_id])



    


import sqlalchemy as sa
import sqlalchemy.orm as so
from sqlalchemy.ext.declarative import declared_attr
from typing import Optional, TYPE_CHECKING, List
from datetime import date
if TYPE_CHECKING:
    from .domain import Domain
    from .vocabulary import Vocabulary
    from .concept_class import Concept_Class
    from .concept_ancestor import Concept_Ancestor
    from .concept_relationship import Concept_Relationship

from omop_alchemy.cdm.base import ReferenceTable, Base, cdm_table, CDMTableBase, ReferenceContextMixin

@cdm_table
class Concept(
    ReferenceTable,
    CDMTableBase,
    Base
):
    __tablename__ = "concept"
    concept_id: so.Mapped[int] = so.mapped_column(primary_key=True)
    concept_name: so.Mapped[str] = so.mapped_column(sa.String(255), nullable=False)
    domain_id: so.Mapped[str] = so.mapped_column(sa.ForeignKey("domain.domain_id"), nullable=False, index=True)
    vocabulary_id: so.Mapped[str] = so.mapped_column(sa.ForeignKey("vocabulary.vocabulary_id"), nullable=False, index=True)
    concept_class_id: so.Mapped[str] = so.mapped_column(sa.ForeignKey("concept_class.concept_class_id"), nullable=False, index=True)
    standard_concept: so.Mapped[Optional[str]] = so.mapped_column(sa.String(1))
    concept_code: so.Mapped[str] = so.mapped_column(sa.String(50), nullable=False)
    valid_start_date: so.Mapped[date] = so.mapped_column(nullable=False)
    valid_end_date: so.Mapped[date] = so.mapped_column(nullable=False)
    invalid_reason: so.Mapped[Optional[str]] = so.mapped_column(sa.String(1))

class ConceptContext(ReferenceContextMixin):
    """
    Navigational relationships for Concept.

    This mixin defines read-only ORM relationships that resolve
    foreign keys into reference tables and hierarchy navigation.
    """
    
    domain: so.Mapped["Domain"] = ReferenceContextMixin._reference_relationship(target="Domain",local_fk="domain_id",remote_pk="domain_id") # type: ignore[assignment]
    vocabulary: so.Mapped["Vocabulary"] = ReferenceContextMixin._reference_relationship(target="Vocabulary",local_fk="vocabulary_id",remote_pk="vocabulary_id") # type: ignore[assignment]
    concept_class: so.Mapped["Concept_Class"] = ReferenceContextMixin._reference_relationship(target="Concept_Class",local_fk="concept_class_id",remote_pk="concept_class_id") # type: ignore[assignment]

    @declared_attr
    def outgoing_relationships(cls) -> so.Mapped[List["Concept_Relationship"]]:
        return so.relationship(
            "Concept_Relationship",
            primaryjoin=f"{cls.__name__}.concept_id == Concept_Relationship.concept_id_1", # type: ignore
            foreign_keys="Concept_Relationship.concept_id_1",
            viewonly=True,
            lazy="selectin",
        )

    @declared_attr
    def incoming_relationships(cls) -> so.Mapped[List["Concept_Relationship"]]:
        return so.relationship(
            "Concept_Relationship",
            primaryjoin=f"{cls.__name__}.concept_id == Concept_Relationship.concept_id_2", # type: ignore
            foreign_keys="Concept_Relationship.concept_id_2",
            viewonly=True,
            lazy="selectin",
        )

    @declared_attr
    def ancestors(cls) -> so.Mapped[List["Concept_Ancestor"]]:
        return so.relationship(
            "Concept_Ancestor",
            primaryjoin=f"{cls.__name__}.concept_id == Concept_Ancestor.descendant_concept_id", # type: ignore
            foreign_keys="Concept_Ancestor.descendant_concept_id",
            viewonly=True,
            lazy="selectin",
        )

    @declared_attr
    def descendants(cls) -> so.Mapped[List["Concept_Ancestor"]]:
        return so.relationship(
            "Concept_Ancestor",
            primaryjoin=f"{cls.__name__}.concept_id == Concept_Ancestor.ancestor_concept_id", # type: ignore
            foreign_keys="Concept_Ancestor.ancestor_concept_id",
            viewonly=True,
            lazy="selectin",
        )

class ConceptView(Concept, ConceptContext):
    """
    Rich, navigable Concept mapping.

    Use when:
    - traversing vocabulary relationships
    - exploring hierarchies
    - semantic inspection

    Avoid in tight loops or ETL paths.
    """
    __tablename__ = "concept"
    __mapper_args__ = {"concrete": False}


    @property
    def is_standard(self) -> bool:
        return self.standard_concept == "S"

    @property
    def is_valid(self) -> bool:
        return self.invalid_reason is None


from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Care_Site(Base):
    __tablename__ = 'care_site'

    # identifier    
    care_site_id: so.Mapped[int] = so.mapped_column(primary_key=True, autoincrement=True) 
    # temporal
    # strings
    care_site_name: so.Mapped[Optional[str]] = so.mapped_column(sa.String(255))
    care_site_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    place_of_service_source_value: so.Mapped[Optional[str]] = so.mapped_column(sa.String(50))
    # numeric
    # fks
    location_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('location.location_id', name='cs_fk_1'))
    # concept fks
    place_of_service_concept_id: so.Mapped[Optional[int]] = so.mapped_column(sa.ForeignKey('concept.concept_id', name='cs_fk_2'))
    # relationships
    location: so.Mapped[Optional['Location']] = so.relationship(foreign_keys=[location_id])    
    # concept relationships
    place_of_service_concept: so.Mapped[Optional['Concept']] = so.relationship(foreign_keys=[place_of_service_concept_id])


from datetime import datetime
from typing import Optional
import sqlalchemy as sa
import sqlalchemy.orm as so

from ...db import Base

class Source_To_Concept_Map(Base):
    __tablename__ = 'source_to_concept_map'
    source_to_concept_map_id: so.Mapped[int] = so.mapped_column(index=True, unique=True, primary_key=True)
